#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import os

from typing import Callable, Sequence, Mapping, MutableMapping, Any
from functools import wraps

from qiling.os.const import *
from qiling.os.windows.utils import *
from qiling.const import *
from qiling.exception import *
from qiling.os.windows.structs import *

from .utils import *

class QlOsFncc:
    def __init__(self, ql):
        self.ql = ql

        # choose a calling convention according to arch and os
        if self.ql.archtype in (QL_ARCH.X86, QL_ARCH.A8086):
            cc = 'cdecl'
        elif self.ql.archtype == QL_ARCH.X8664:
            cc = {
                QL_OS.LINUX:   'amd64',
                QL_OS.MACOS:   'macosx64',
                QL_OS.WINDOWS: 'ms',
                QL_OS.UEFI:    'ms'
            }.get(self.ql.ostype, '')
        elif self.ql.archtype == QL_ARCH.MIPS:
            cc = {
                QL_OS.LINUX:   'mips_o32'
            }.get(self.ql.ostype, '')                        
        else:
            # do not pick a cc; let class overrides define the necessary handlers
            cc = ''

        # register used to pass the return value to the caller
        self.__cc_reg_ret = {
            QL_ARCH.A8086: UC_X86_REG_AX,
            QL_ARCH.X86:   UC_X86_REG_EAX,
            QL_ARCH.X8664: UC_X86_REG_RAX,
            QL_ARCH.MIPS: UC_MIPS_REG_2
        }.get(self.ql.archtype, UC_X86_REG_INVALID)

        # registers used to pass arguments; a None stands for a stack argument
        self._cc_args = {
            'amd64': (UC_X86_REG_RDI, UC_X86_REG_RSI, UC_X86_REG_RDX, UC_X86_REG_R10, UC_X86_REG_R8, UC_X86_REG_R9) + (None, ) * 10,
            'macosx64': (UC_X86_REG_RDI, UC_X86_REG_RSI, UC_X86_REG_RDX, UC_X86_REG_RCX, UC_X86_REG_R8, UC_X86_REG_R9) + (None, ) * 10,
            'cdecl': (None, ) * 16,
            'ms':    (UC_X86_REG_RCX, UC_X86_REG_RDX, UC_X86_REG_R8, UC_X86_REG_R9) + (None, ) * 12,
            'mips_o32':  (UC_MIPS_REG_4, UC_MIPS_REG_5, UC_MIPS_REG_6, UC_MIPS_REG_7) + (None, ) * 31
        }.get(cc, [])

        # shadow stack size in terms of stack items
        self._shadow = 4 if cc == 'ms' else 0

        # arch native address size in bytes
        self._asize = self.ql.archbit // 8

    def get_param_by_index(self, index: int) -> int:
        """Get an argument value by its index.
        """
        max_args = len(self._cc_args)
        assert index < max_args, f'currently supporting up to {max_args} args'

        reg = self._cc_args[index]

        # should arg be read from a reg or the stack?
        if reg is None:
            if self.ql.archtype == QL_ARCH.MIPS:
                return self.ql.stack.read(index * self._asize)

            # get matching stack item
            si = index - self._cc_args.index(None)

            # skip return address and shadow space
            return self.ql.stack_read((1 + self._shadow + si) * self._asize)
        else:
            return self.ql.uc.reg_read(reg)


    def get_function_param(self, nargs: int) -> list:
        """Get values of `nargs` first arguments.
        """

        params = [self.get_param_by_index(i) for i in range(nargs)]

        return params[0] if nargs == 1 else params

    def set_function_params(self, in_params: Mapping[str, int], out_params: MutableMapping[str, Any]) -> int:
        """Retrieve parameters values according to their assigned type.

        Args:
            in_params : a mapping of parameter names to their types
            out_params: a mapping of parameter names to their values
        Returns: number of consumed arguments; this is only relevant for the cdecl
        calling convention, where all the arguments are stored on the stack. in that
        case, the value reflects the number of items consumed from the stack
        """

        def __nullptr_or_deref(idx: int, deref: Callable[[int], Any]):
            ptr = self.get_param_by_index(idx)

            return deref(ptr) if ptr else 0

        def __handle_default(idx: int):
            return self.get_param_by_index(idx), 1

        def __handle_POINTER(idx: int):
            return __handle_default(idx)

        def __handle_ULONGLONG_32(idx: int):
            lo = self.get_param_by_index(idx)
            hi = self.get_param_by_index(idx + 1)

            return (hi << 32) | lo, 2

        def __handle_STRING(idx: int):
            return __nullptr_or_deref(idx, self.read_cstring), 1

        def __handle_WSTRING(idx: int):
            return __nullptr_or_deref(idx, self.read_wstring), 1

        def __handle_GUID(idx: int):
            return __nullptr_or_deref(idx, lambda p: str(self.read_guid(p))), 1

        param_handlers = {
            ULONGLONG: __handle_POINTER if self.ql.archbit == 64 else __handle_ULONGLONG_32,
            POINTER  : __handle_POINTER,
            STRING   : __handle_STRING,
            WSTRING  : __handle_WSTRING,
            GUID     : __handle_GUID
        }

        i = 0
        for pname, ptype in in_params.items():
            handler = param_handlers.get(ptype, __handle_default)
            out_params[pname], consumed = handler(i)
            i += consumed

        return i

    def set_return_value(self, ret: int):
        self.ql.uc.reg_write(self.__cc_reg_ret, ret)

    def get_return_value(self) -> int:
        return self.ql.uc.reg_read(self.__cc_reg_ret)

    #
    # stdcall cdecl fastcall cc
    #
    def __cc(self, param_num, params, func, args, kwargs, passthru=False):
        # read params values
        if params is not None:
            param_num = self.set_function_params(params, args[2])


        if self.ql.ostype in (QL_OS.WINDOWS, QL_OS.UEFI):
            # if set, fire up the on-enter hook
            if callable(self.winapi_func_onenter):
                address, params = self.winapi_func_onenter(*args, **kwargs)

                # override original args set
                args = (self.ql, address, params)

        # call function
        result = func(*args, **kwargs)

        if self.ql.ostype in (QL_OS.WINDOWS, QL_OS.UEFI):
            # if set, fire up the on-exit hook
            if callable(self.winapi_func_onexit):
                self.winapi_func_onexit(*args, **kwargs)

        # set return value
        if result is not None:
            self.set_return_value(result)

        # print
        self.print_function(args[1], func.__name__, args[2], result, passthru)

        return result, param_num

    def x86_stdcall(self, param_num, params, func, args, kwargs, passthru=False):
        # if we check ret_addr before the call, we can't modify the ret_addr from inside the hook
        result, param_num = self.__cc(param_num, params, func, args, kwargs, passthru)

        ret_addr = self.ql.stack_read(0)

        # append syscall to list
        if self.ql.ostype in (QL_OS.WINDOWS, QL_OS.UEFI):
            self._call_api(func.__name__, params, result, self.ql.reg.arch_pc, ret_addr)
            if not self.PE_RUN:
                return result
        
        if not passthru:
            # callee is responsible for cleaning up the stack; unwind the stack
            self.ql.reg.arch_sp = self.ql.reg.arch_sp + ((param_num + 1) * self._asize)
            self.ql.reg.arch_pc = ret_addr

        return result

    def x86_cdecl(self, param_num, params, func, args, kwargs, passthru=False):
        result, _ = self.__cc(param_num, params, func, args, kwargs)
        old_pc = self.ql.reg.arch_pc

        # append syscall to list
        if self.ql.ostype in (QL_OS.WINDOWS, QL_OS.UEFI):
            self._call_api(func.__name__, params, result, old_pc, self.ql.stack_read(0))
            if not self.PE_RUN:
                return result

        if not passthru:
            self.ql.reg.arch_pc = self.ql.stack_pop()

        return result

    def x8664_fastcall(self, param_num, params, func, args, kwargs, passthru=False):
        result, param_num = self.__cc(param_num, params, func, args, kwargs)
        old_pc = self.ql.reg.arch_pc

        # append syscall to list
        if self.ql.ostype in (QL_OS.WINDOWS, QL_OS.UEFI):
            self._call_api(func.__name__, params, result, old_pc, self.ql.stack_read(0))
            if not self.PE_RUN:
                return result

        if not passthru:
            self.ql.reg.arch_pc = self.ql.stack_pop()
        
        return result

    def mips_o32_call(self, param_num, params, func, args, kwargs):
        result, param_num  = self.__cc(param_num, params, func, args, kwargs)
        self.ql.reg.arch_pc = self.ql.reg.ra

        return result


#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import os

from qiling.os.thread import *
from qiling.os.macos.mach_port import *

# TODO: finished
class QlMachoThread(QlThread):

    def __init__(self, ql):
        super(QlMachoThread, self).__init__(ql)
        self.ql = ql
        self.id = 123
        self.port = MachPort(0x307)

    def stop(self):
        pass


class QlMachoThreadManagement:

    def __init__(self, ql):
        self.ql = ql
        self.cur_thread = None



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import struct
import time
from datetime import datetime
from qiling.os.windows.const import *
from qiling.os.const import *
from qiling.os.windows.fncc import *
from qiling.os.windows.utils import *
from qiling.os.windows.thread import *
from qiling.os.windows.handle import *
from qiling.exception import *
from qiling.os.windows.structs import *

dllname = 'kernel32_dll'

# NOT_BUILD_WINDOWS_DEPRECATE DWORD GetVersion(
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetVersion(ql, address, params):
    ret = 0x0004
    ret = ret << 16
    ret = ret | 0x0004
    return ret


# NOT_BUILD_WINDOWS_DEPRECATE BOOL GetVersionExA(
#   LPOSVERSIONINFOA lpVersionInformation
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetVersionExA(ql, address, params):
    return hook_GetVersionExW.__wrapped__(ql, address, params)


# NOT_BUILD_WINDOWS_DEPRECATE BOOL GetVersionExW(
#   LPOSVERSIONINFOW lpVersionInformation
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetVersionExW(ql, address, params):
    return 1


# void GetSystemInfo(
#   LPSYSTEM_INFO lpSystemInfo
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetSystemInfo(ql, address, params):
    pointer = params["lpSystemInfo"]
    system_info = SystemInfo(ql, 0, ql.os.heap.page_size, ql.loader.pe_image_address,
                             ql.loader.dll_address + ql.loader.dll_size, 0x3, 0x4, 0x24a, ql.os.heap.page_size * 10,
                             0x6, 0x4601)
    system_info.write(pointer)
    return 0


# void GetLocalTime(
#   LPSYSTEMTIME lpSystemTime
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetLocalTime(ql, address, params):
    import datetime
    ptr = params['lpSystemTime']
    d = datetime.datetime.now()
    system_time = SystemTime(ql, d.year, d.month, d.isoweekday(), d.day, d.hour, d.minute, d.second,
                             d.microsecond // 1000)
    system_time.write(ptr)
    return 0


# void GetSystemTimeAsFileTime(
#   LPFILETIME lpSystemTimeAsFileTime
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetSystemTimeAsFileTime(ql, address, params):
    # TODO
    pass


# DWORD GetTickCount(
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetTickCount(ql, address, params):
    ret = 200000
    return ret


# UINT GetWindowsDirectoryW(
#   LPWSTR lpBuffer,
#   UINT   uSize
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetWindowsDirectoryW(ql, address, params):
    dst = params["lpBuffer"]
    value = (ql.os.windir + "\x00").encode("utf-16le")
    ql.mem.write(dst, value)
    return len(value) - 2

# UINT GetSystemWindowsDirectoryW(
#   LPWSTR lpBuffer,
#   UINT   uSize
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetSystemWindowsDirectoryW(ql, address, params):
    return hook_GetWindowsDirectoryW.__wrapped__(ql, address, params)

# void GetNativeSystemInfo(
#   LPSYSTEM_INFO lpSystemInfo
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetNativeSystemInfo(ql, address, params):
    pointer = params["lpSystemInfo"]
    system_info = SystemInfo(ql, 0, ql.os.heap.page_size, ql.loader.pe_image_address,
                             ql.loader.dll_address + ql.loader.dll_size, 0x3, 0x4, 0x24a, ql.os.heap.page_size * 10,
                             0x6, 0x4601)
    system_info.write(pointer)
    return 0

# void GetSystemTime(
#   LPSYSTEMTIME lpSystemTime);
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetSystemTIme(ql, address, params):
    dt = datetime.now().microsecond.to_bytes(8, byteorder="little")
    pointer = params["lpSystemTime"]
    ql.mem.write(pointer, dt)
    return 0


# typedef struct _FILETIME {
#   DWORD dwLowDateTime;
#   DWORD dwHighDateTime;
# } FILETIME, *PFILETIME, *LPFILETIME;

# void GetSystemTimePreciseAsFileTime(
#   LPFILETIME lpSystemTimeAsFileTime
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetSystemTimePreciseAsFileTime(ql, address, params):
    # todo check if the value is correct
    dt = datetime.now().microsecond.to_bytes(8, byteorder="little")
    pointer = params["lpSystemTimeAsFileTime"]
    ql.mem.write(pointer, dt)
    return 0



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import struct
import time
from qiling.os.windows.const import *
from qiling.os.const import *
from qiling.os.windows.fncc import *
from qiling.os.windows.utils import *
from qiling.os.windows.thread import *
from qiling.os.windows.handle import *
from qiling.exception import *

dllname = 'ole32_dll'

# HRESULT OleInitialize(
#   IN LPVOID pvReserved
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_OleInitialize(ql, address, params):
    # I don't think we need to do anything, we hook every call for the COM library and manage them locally
    return S_OK


# HRESULT CoRegisterMessageFilter(
#   LPMESSAGEFILTER lpMessageFilter,
#   LPMESSAGEFILTER *lplpMessageFilter
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_CoRegisterMessageFilter(ql, address, params):
    return S_OK


# HRESULT CoInitializeEx(
#   LPVOID pvReserved,
#   DWORD  dwCoInit
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_CoInitializeEx(ql, address, params):
    return S_OK

# HRESULT CoInitializeSecurity(
#   PSECURITY_DESCRIPTOR        pSecDesc,
#   LONG                        cAuthSvc,
#   SOLE_AUTHENTICATION_SERVICE *asAuthSvc,
#   void                        *pReserved1,
#   DWORD                       dwAuthnLevel,
#   DWORD                       dwImpLevel,
#   void                        *pAuthList,
#   DWORD                       dwCapabilities,
#   void                        *pReserved3
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_CoInitializeSecurity(ql, address, params):
    return S_OK


# HRESULT CoCreateInstance(
#   REFCLSID  rclsid,
#   LPUNKNOWN pUnkOuter,
#   DWORD     dwClsContext,
#   REFIID    riid,
#   LPVOID    *ppv
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_CoCreateInstance(ql, address, params):
    # FIXME: probably this needs implementation
    return S_OK


#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from abc import ABC, abstractmethod


class QlBaseCoverage(ABC):
    """
    An abstract base class for concrete code coverage collectors.
    To add support for a new coverage format, just derive from this class and implement
    all the methods marked with the @abstractmethod decorator.
    """
    
    def __init__(self):
        super().__init__()

    @property
    @staticmethod
    @abstractmethod
    def FORMAT_NAME():
        raise NotImplementedError

    @abstractmethod
    def activate(self):
        pass

    @abstractmethod
    def deactivate(self):
        pass

    @abstractmethod
    def dump_coverage(self, coverage_file):
        pass




#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#
import struct
import base64
from qiling.os.windows.fncc import *
from qiling.os.const import *
from qiling.os.windows.utils import *
from qiling.os.windows.handle import *
from qiling.os.windows.const import *

dllname = 'mscoree_dll'

# void STDMETHODCALLTYPE CorExitProcess (
#   int  exitCode
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'int': 'DWORD'})
def hook_CorExitProcess(ql, address, params):
    ql.emu_stop()
    ql.os.PE_RUN = False


# __int32 STDMETHODCALLTYPE _CorExeMain ();
@winsdkapi(cc=STDCALL, dllname='crypt32_dll')
def hook__CorExeMain(ql, address, params):
    # TODO implement + check call type
    pass



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import math, copy, os
from contextlib import contextmanager

from qiling.const import QL_ARCH
from .utils import dump_regs, get_arm_flags, disasm
from .const import color


# read data from memory of qiling instance
def examine_mem(ql, addr, fmt):

    def unpack(bs, sz):
        return {
                1: lambda x: x[0],
                2: ql.unpack16,
                4: ql.unpack32,
                8: ql.unpack64,
                }.get(sz)(bs)

    ft, sz, ct = fmt

    if ft == "i":

        for offset in range(addr, addr+ct*4, 4):
            line = disasm(ql, offset)
            if line:
                print("0x{:x}: {}\t{}".format(line.address, line.mnemonic, line.op_str))

        print()

    else:
        lines = 1 if ct <= 4 else math.ceil(ct / 4)

        mem_read = [ql.mem.read(addr+(offset*sz), sz) for offset in range(ct)]

        for line in range(lines):
            offset = line * sz * 4
            print("0x{:x}:\t".format(addr+offset), end="")

            idx = line * 4
            for each in mem_read[idx:idx+4]:
                data = unpack(each, sz)
                prefix = "0x" if ft in ("x", "a") else ""
                pad = '0' + str(sz*2) if ft in ('x', 'a', 't') else ''
                ft = ft.lower() if ft in ("x", "o", "b", "d") else ft.lower().replace("t", "b").replace("a", "x")

                print("{}{{:{}{}}}\t".format(prefix, pad, ft).format(data), end="")

            print()


# get terminal window height and width
def get_terminal_size():
    return map(int, os.popen('stty size', 'r').read().split())


# try to read data from ql memory
def _try_read(ql, address, size):
    try:
        result = ql.mem.read(address, size)
    except:
        result = None

    return result


# divider printer
@contextmanager
def context_printer(ql, field_name, ruler="="):
    _height, _width = get_terminal_size()
    print(field_name, ruler * (_width - len(field_name) - 1))
    yield
    print(ruler * _width)


def context_reg(ql, saved_states=None, *args, **kwargs):

    # context render for registers
    with context_printer(ql, "[Registers]"):

        _cur_regs = dump_regs(ql)

        _colors = (color.DARKCYAN, color.BLUE, color.RED, color.YELLOW, color.GREEN, color.PURPLE, color.CYAN, color.WHITE)

        if ql.archtype == QL_ARCH.MIPS:

            _cur_regs.update({"fp": _cur_regs.pop("s8")})

            if saved_states is not None:
                _saved_states = copy.deepcopy(saved_states)
                _saved_states.update({"fp": _saved_states.pop("s8")})
                _diff = [k for k in _cur_regs if _cur_regs[k] != _saved_states[k]]

            else:
                _diff = None

            lines = ""
            for idx, r in enumerate(_cur_regs, 1):
                line = "{}{}: 0x{{:08x}} {}\t".format(_colors[(idx-1) // 4], r, color.END)

                if _diff and r in _diff:
                    line = "{}{}".format(color.UNDERLINE, color.BOLD) + line

                if idx % 4 == 0 and idx != 32:
                    line += "\n"

                lines += line

            print(lines.format(*_cur_regs.values()))

        elif ql.archtype in (QL_ARCH.ARM, QL_ARCH.ARM_THUMB):

            _cur_regs.update({"sl": _cur_regs.pop("r10")})
            _cur_regs.update({"fp": _cur_regs.pop("r11")})
            _cur_regs.update({"ip": _cur_regs.pop("r12")})

            if saved_states is not None:
                _saved_states = copy.deepcopy(saved_states)
                _saved_states.update({"sl": _saved_states.pop("r10")})
                _saved_states.update({"fp": _saved_states.pop("r11")})
                _saved_states.update({"ip": _saved_states.pop("r12")})
                _diff = [k for k in _cur_regs if _cur_regs[k] != _saved_states[k]]

            else:
                _diff = None

            lines = ""
            for idx, r in enumerate(_cur_regs, 1):
                line = "{}{:}: 0x{{:08x}} {}\t".format(_colors[(idx-1) // 4], r, color.END)

                if _diff and r in _diff:
                    line = "{}{}".format(color.UNDERLINE, color.BOLD) + line

                if idx % 4 == 0:
                    line += "\n"

                lines += line

            print(lines.format(*_cur_regs.values()))
            print(color.GREEN, "[{cpsr[mode]} mode], Thumb: {cpsr[thumb]}, FIQ: {cpsr[fiq]}, IRQ: {cpsr[irq]}, NEG: {cpsr[neg]}, ZERO: {cpsr[zero]}, Carry: {cpsr[carry]}, Overflow: {cpsr[overflow]}".format(cpsr=get_arm_flags(ql.reg.cpsr)), color.END, sep="")

    # context render for Stack
    with context_printer(ql, "[Stack]", ruler="-"):

        for idx in range(8):
            _addr = ql.reg.arch_sp + idx * 4
            _val = ql.mem.read(_addr, ql.archbit // 8)
            print("$sp+0x%02x|[0x%08x]=> 0x%08x" % (idx*4, _addr, ql.unpack(_val)), end="")

            try: # try to deference wether its a pointer
                _deref = ql.mem.read(_addr, 4)
            except:
                _deref = None

            if _deref:
                print(" => 0x%08x" % ql.unpack(_deref))


def print_asm(ql, instructions):
    for ins in instructions:
        fmt = (ins.address, ins.mnemonic.ljust(6), ins.op_str)
        if ql.reg.arch_pc == ins.address:
            print("PC ==>  0x%x\t%s %s" % fmt)
        else:
            print("\t0x%x\t%s %s" % fmt)


def context_asm(ql, address, size, *args, **kwargs):

    with context_printer(ql, field_name="[Code]"):
        md = ql.create_disassembler()

        # assembly before current location

        pre_tmp = _try_read(ql, address-0x10, 0x10)
        if pre_tmp:
            pre_ins = md.disasm(pre_tmp, address-0x10)
            print_asm(ql, pre_ins)

        # assembly for current locaton

        tmp = ql.mem.read(address, size)
        cur_ins = md.disasm(tmp, address)
        print_asm(ql, cur_ins)

        # assembly after current locaton

        pos_tmp = _try_read(ql, address+4, 0x10)
        if pos_tmp:
            pos_ins = md.disasm(pos_tmp, address+4)
            print_asm(ql, pos_ins)



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import json

from qiling.const import *


class Report:
    def __init__(self, ql):
        self.filename = ql.argv
        self.rootfs = ql.rootfs
        self.arch = list(arch_map.keys())[list(arch_map.values()).index(ql.archtype)]
        self.os = list(os_map.keys())[list(os_map.values()).index(ql.ostype)]
        self.env = ql.env
        self.strings = set()
        for string in ql.os.appeared_strings:
            strings = string.split(" ")
            self.strings |= set(strings)
        self.profile = {}
        for section in ql.profile.sections():
            self.profile.update(dict(ql.profile.items(section)))
        self.strings = list(self.strings)
        self.patches = []
        self.patches.extend(ql.patch_bin)
        self.patches.extend(ql.patch_lib)


class WindowsReport(Report):
    def __init__(self, ql):
        super().__init__(ql)
        self.dlls = ql.loader.dlls
        self.teb_address = ql.loader.TEB.base
        self.peb_address = ql.loader.PEB.base
        self.ldr_address = ql.loader.LDR.base
        self.api = ql.os.syscalls
        self.registries = {}
        for key, values in ql.os.registry_manager.accessed.items():
            self.registries[key] = values
        self.entry_point = ql.loader.entry_point
        self.import_symbols = {}
        for key, internal_dict in ql.loader.import_symbols.items():
            internal_dict["name"] = internal_dict["name"].decode("utf-8") if type(internal_dict["name"]) == bytes else \
                                    internal_dict["name"]
            self.import_symbols[key] = internal_dict
        self.export_symbols = {}
        for key, internal_dict in ql.loader.export_symbols.items():
            internal_dict["name"] = internal_dict["name"].decode("utf-8") if type(internal_dict["name"]) == bytes else \
                                    internal_dict["name"]
            self.export_symbols[key] = internal_dict
        self.cmdline = ql.loader.cmdline.decode("utf-8")


def generate_report(ql, pretty_print=False) -> dict:
    if ql.ostype == QL_OS.WINDOWS:
        report = WindowsReport(ql)
    else:
        report = Report(ql)
    res = report.__dict__
    if pretty_print:
        res = json.dumps(res, indent=10, sort_keys=True)
    return res



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from qiling.os.const import *
from ..fncc import *
from ..ProcessorBind import *
from ..UefiBaseType import *

PCD_PROTOCOL_CALLBACK = FUNCPTR(VOID, PTR(EFI_GUID), UINTN, PTR(VOID), UINTN)

PCD_PROTOCOL_SET_SKU		= FUNCPTR(VOID, UINTN)
PCD_PROTOCOL_GET8			= FUNCPTR(UINT8, UINTN)
PCD_PROTOCOL_GET16			= FUNCPTR(UINT16, UINTN)
PCD_PROTOCOL_GET32			= FUNCPTR(UINT32, UINTN)
PCD_PROTOCOL_GET64			= FUNCPTR(UINT64, UINTN)
PCD_PROTOCOL_GET_POINTER	= FUNCPTR(PTR(VOID), UINTN)
PCD_PROTOCOL_GET_BOOLEAN	= FUNCPTR(BOOLEAN, UINTN)
PCD_PROTOCOL_GET_SIZE		= FUNCPTR(UINTN, UINTN)
PCD_PROTOCOL_GET_EX_8		= FUNCPTR(UINT8, PTR(EFI_GUID), UINTN)
PCD_PROTOCOL_GET_EX_16		= FUNCPTR(UINT16, PTR(EFI_GUID), UINTN)
PCD_PROTOCOL_GET_EX_32		= FUNCPTR(UINT32, PTR(EFI_GUID), UINTN)
PCD_PROTOCOL_GET_EX_64		= FUNCPTR(UINT64, PTR(EFI_GUID), UINTN)
PCD_PROTOCOL_GET_EX_POINTER	= FUNCPTR(PTR(VOID), PTR(EFI_GUID), UINTN)
PCD_PROTOCOL_GET_EX_BOOLEAN	= FUNCPTR(BOOLEAN, PTR(EFI_GUID), UINTN)
PCD_PROTOCOL_GET_EX_SIZE	= FUNCPTR(UINTN, PTR(EFI_GUID), UINTN)
PCD_PROTOCOL_SET8			= FUNCPTR(EFI_STATUS, UINTN, UINT8)
PCD_PROTOCOL_SET16			= FUNCPTR(EFI_STATUS, UINTN, UINT16)
PCD_PROTOCOL_SET32			= FUNCPTR(EFI_STATUS, UINTN, UINT32)
PCD_PROTOCOL_SET64			= FUNCPTR(EFI_STATUS, UINTN, UINT64)
PCD_PROTOCOL_SET_POINTER	= FUNCPTR(EFI_STATUS, UINTN, PTR(UINTN), PTR(VOID))
PCD_PROTOCOL_SET_BOOLEAN	= FUNCPTR(EFI_STATUS, UINTN, BOOLEAN)
PCD_PROTOCOL_SET_EX_8		= FUNCPTR(EFI_STATUS, PTR(EFI_GUID), UINTN, UINT8)
PCD_PROTOCOL_SET_EX_16		= FUNCPTR(EFI_STATUS, PTR(EFI_GUID), UINTN, UINT16)
PCD_PROTOCOL_SET_EX_32		= FUNCPTR(EFI_STATUS, PTR(EFI_GUID), UINTN, UINT32)
PCD_PROTOCOL_SET_EX_64		= FUNCPTR(EFI_STATUS, PTR(EFI_GUID), UINTN, UINT64)
PCD_PROTOCOL_SET_EX_POINTER	= FUNCPTR(EFI_STATUS, PTR(EFI_GUID), UINTN, PTR(UINTN), PTR(VOID))
PCD_PROTOCOL_SET_EX_BOOLEAN	= FUNCPTR(EFI_STATUS, PTR(EFI_GUID), UINTN, BOOLEAN)
PCD_PROTOCOL_CALLBACK_ONSET	= FUNCPTR(EFI_STATUS, PTR(EFI_GUID), UINTN, PCD_PROTOCOL_CALLBACK)
PCD_PROTOCOL_CANCEL_CALLBACK= FUNCPTR(EFI_STATUS, PTR(EFI_GUID), UINTN, PCD_PROTOCOL_CALLBACK)
PCD_PROTOCOL_GET_NEXT_TOKEN	= FUNCPTR(EFI_STATUS, PTR(EFI_GUID), PTR(UINTN))
PCD_PROTOCOL_GET_NEXT_TOKENSPACE = FUNCPTR(EFI_STATUS, PTR(PTR(EFI_GUID)))

class PCD_PROTOCOL(STRUCT):
	_fields_ = [
		('SetSku',				PCD_PROTOCOL_SET_SKU),
		('Get8',				PCD_PROTOCOL_GET8),
		('Get16',				PCD_PROTOCOL_GET16),
		('Get32',				PCD_PROTOCOL_GET32),
		('Get64',				PCD_PROTOCOL_GET64),
		('GetPtr',				PCD_PROTOCOL_GET_POINTER),
		('GetBool',				PCD_PROTOCOL_GET_BOOLEAN),
		('GetSize',				PCD_PROTOCOL_GET_SIZE),
		('Get8Ex',				PCD_PROTOCOL_GET_EX_8),
		('Get16Ex',				PCD_PROTOCOL_GET_EX_16),
		('Get32Ex',				PCD_PROTOCOL_GET_EX_32),
		('Get64Ex',				PCD_PROTOCOL_GET_EX_64),
		('GetPtrEx',			PCD_PROTOCOL_GET_EX_POINTER),
		('GetBoolEx',			PCD_PROTOCOL_GET_EX_BOOLEAN),
		('GetSizeEx',			PCD_PROTOCOL_GET_EX_SIZE),
		('Set8',				PCD_PROTOCOL_SET8),
		('Set16',				PCD_PROTOCOL_SET16),
		('Set32',				PCD_PROTOCOL_SET32),
		('Set64',				PCD_PROTOCOL_SET64),
		('SetPtr',				PCD_PROTOCOL_SET_POINTER),
		('SetBool',				PCD_PROTOCOL_SET_BOOLEAN),
		('Set8Ex',				PCD_PROTOCOL_SET_EX_8),
		('Set16Ex',				PCD_PROTOCOL_SET_EX_16),
		('Set32Ex',				PCD_PROTOCOL_SET_EX_32),
		('Set64Ex',				PCD_PROTOCOL_SET_EX_64),
		('SetPtrEx',			PCD_PROTOCOL_SET_EX_POINTER),
		('SetBoolEx',			PCD_PROTOCOL_SET_EX_BOOLEAN),
		('CallbackOnSet',		PCD_PROTOCOL_CALLBACK_ONSET),
		('CancelCallback',		PCD_PROTOCOL_CANCEL_CALLBACK),
		('GetNextToken',		PCD_PROTOCOL_GET_NEXT_TOKEN),
		('GetNextTokenSpace',	PCD_PROTOCOL_GET_NEXT_TOKENSPACE)
	]

@dxeapi(params = {
	"SkuId" : UINT
})
def hook_SetSku(ql, address, params):
	pass

@dxeapi(params = {
	"TokenNumber" : UINT
})
def hook_Get8(ql, address, params):
	pass

@dxeapi(params = {
	"TokenNumber" : UINT
})
def hook_Get16(ql, address, params):
	pass

@dxeapi(params = {
	"TokenNumber" : UINT
})
def hook_Get32(ql, address, params):
	pass

@dxeapi(params = {
	"TokenNumber" : UINT
})
def hook_Get64(ql, address, params):
	pass

@dxeapi(params = {
	"TokenNumber" : UINT
})
def hook_GetPtr(ql, address, params):
	pass

@dxeapi(params = {
	"TokenNumber" : UINT
})
def hook_GetBool(ql, address, params):
	pass

@dxeapi(params = {
	"TokenNumber" : UINT
})
def hook_GetSize(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: UINT
})
def hook_Get8Ex(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: UINT
})
def hook_Get16Ex(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: UINT
})
def hook_Get32Ex(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: UINT
})
def hook_Get64Ex(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: UINT
})
def hook_GetPtrEx(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: UINT
})
def hook_GetBoolEx(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: UINT
})
def hook_GetSizeEx(ql, address, params):
	pass

@dxeapi(params = {
	"TokenNumber"	: UINT,
	"Value"			: INT
})
def hook_Set8(ql, address, params):
	pass

@dxeapi(params = {
	"TokenNumber"	: UINT,
	"Value"			: INT
})
def hook_Set16(ql, address, params):
	pass

@dxeapi(params = {
	"TokenNumber"	: UINT,
	"Value"			: INT
})
def hook_Set32(ql, address, params):
	pass

@dxeapi(params = {
	"TokenNumber"	: UINT,
	"Value"			: INT
})
def hook_Set64(ql, address, params):
	pass

@dxeapi(params = {
	"TokenNumber"	: UINT,
	"SizeOfValue"	: POINTER,
	"Buffer"		: POINTER
})
def hook_SetPtr(ql, address, params):
	pass

@dxeapi(params = {
	"TokenNumber"	: UINT,
	"Value"			: INT
})
def hook_SetBool(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: UINT,
	"Value"			: INT
})
def hook_Set8Ex(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: UINT,
	"Value"			: INT
})
def hook_Set16Ex(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: UINT,
	"Value"			: INT
})
def hook_Set32Ex(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: UINT,
	"Value"			: INT
})
def hook_Set64Ex(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: UINT,
	"SizeOfValue"	: POINTER,
	"Buffer"		: POINTER
})
def hook_SetPtrEx(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: UINT,
	"Value"			: INT
})
def hook_SetBoolEx(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"				: GUID,
	"TokenNumber"		: UINT,
	"CallBackFunction"	: POINTER
})
def hook_CallbackOnSet(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"				: GUID,
	"TokenNumber"		: UINT,
	"CallBackFunction"	: POINTER
})
def hook_CancelCallback(ql, address, params):
	pass

@dxeapi(params = {
	"Guid"			: GUID,
	"TokenNumber"	: POINTER
})
def hook_GetNextToken(ql, address, params):
	pass

@dxeapi(params = {
	"Guid" : POINTER
})
def hook_GetNextTokenSpace(ql, address, params):
	pass

descriptor = {
	"guid" : "11b34006-d85b-4d0a-a290-d5a571310ef7",
	"struct" : PCD_PROTOCOL,
	"fields" : (
		('SetSku',				hook_SetSku),
		('Get8',				hook_Get8),
		('Get16',				hook_Get16),
		('Get32',				hook_Get32),
		('Get64',				hook_Get64),
		('GetPtr',				hook_GetPtr),
		('GetBool',				hook_GetBool),
		('GetSize',				hook_GetSize),
		('Get8Ex',				hook_Get8Ex),
		('Get16Ex',				hook_Get16Ex),
		('Get32Ex',				hook_Get32Ex),
		('Get64Ex',				hook_Get64Ex),
		('GetPtrEx',			hook_GetPtrEx),
		('GetBoolEx',			hook_GetBoolEx),
		('GetSizeEx',			hook_GetSizeEx),
		('Set8',				hook_Set8),
		('Set16',				hook_Set16),
		('Set32',				hook_Set32),
		('Set64',				hook_Set64),
		('SetPtr',				hook_SetPtr),
		('SetBool',				hook_SetBool),
		('Set8Ex',				hook_Set8Ex),
		('Set16Ex',				hook_Set16Ex),
		('Set32Ex',				hook_Set32Ex),
		('Set64Ex',				hook_Set64Ex),
		('SetPtrEx',			hook_SetPtrEx),
		('SetBoolEx',			hook_SetBoolEx),
		('CallbackOnSet',		hook_CallbackOnSet),
		('CancelCallback',		hook_CancelCallback),
		('GetNextToken',		hook_GetNextToken),
		('GetNextTokenSpace',	hook_GetNextTokenSpace)
	)
}



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from unicorn import UcError

from qiling.os.os import QlOs
from qiling.os.fncc import QlOsFncc

class QlOsUefi(QlOs, QlOsFncc):
	def __init__(self, ql):
		QlOs.__init__(self, ql)
		QlOsFncc.__init__(self, ql)
		self.ql = ql
		self.entry_point = 0
		self.running_module = None
		self.user_defined_api = {}
		self.user_defined_api_onenter = {}
		self.user_defined_api_onexit = {}
		self.PE_RUN = True
		self.heap = None # Will be initialized by the loader.


	def save(self):
		saved_state = super(QlOsUefi, self).save()
		saved_state['entry_point'] = self.entry_point
		return saved_state


	def restore(self, saved_state):
		super(QlOsUefi, self).restore(saved_state)
		self.entry_point = saved_state['entry_point']


	@staticmethod
	def notify_after_module_execution(ql, number_of_modules_left):
		return False


	@staticmethod
	def notify_before_module_execution(ql, module):
		ql.os.running_module = module
		return False


	def emit_context(self):
		# TODO: add xmm, ymm, zmm registers
		rgroups = (
			('rax', 'eax', 'ax', 'ah', 'al'),
			('rbx', 'ebx', 'bx', 'bh', 'bl'),
			('rcx', 'ecx', 'cx', 'ch', 'cl'),
			('rdx', 'edx', 'dx', 'dh', 'dl'),
			('rsi', 'esi', 'si', ''),  # BUG: sil is missing
			('rdi', 'edi', 'di', ''),  # BUG: dil is missing
			('rsp', 'esp', 'sp', ''),  # BUG: spl is missing
			('rbp', 'ebp', 'bp', ''),  # BUG: bpl is missing
			('rip', 'eip', 'ip', ''),
			(),
			('r8',  'r8d',  'r8w',  'r8b' ),
			('r9',  'r9d',  'r9w',  'r9b' ),
			('r10', 'r10d', 'r10w', 'r10b'),
			('r11', 'r11d', 'r11w', 'r11b'),
			('r12', 'r12d', 'r12w', 'r12b'),
			('r13', 'r13d', 'r13w', 'r13b'),
			('r14', 'r14d', 'r14w', 'r14b'),
			('r15', 'r15d', 'r15w', 'r15b'),
			(),
			('', '', 'cs'),
			('', '', 'ds'),
			('', '', 'es'),
			('', '', 'fs'),
			('', '', 'gs'),
			('', '', 'ss')
		)

		sizes = (64, 32, 16, 8, 8)

		self.ql.log.error(f'CPU Context:')

		for grp in rgroups:
			self.ql.log.error(', '.join((f'{reg:4s} = {self.ql.reg.read(reg):0{bits // 4}x}') for reg, bits in zip(grp, sizes) if reg))

		self.ql.log.error(f'')


	def emit_hexdump(self, address: int, data: bytearray, num_cols: int = 16):
		self.ql.log.error('Hexdump:')

		# align hexdump to numbers of columns
		pre_padding = [None] * (address % num_cols)
		post_padding = [None] * (num_cols - len(pre_padding))
		chars = pre_padding + list(data) + post_padding
		address = address & ~(num_cols - 1)

		for i in range(0, len(chars), num_cols):
			hexdump = ' '.join(f'  ' if ch is None else f'{ch:02x}' for ch in chars[i: i + num_cols])
			self.ql.log.error(f'{address + i:08x} : {hexdump}')

		self.ql.log.error(f'')


	def emit_disasm(self, address: int, data: bytearray, num_insns: int = 8):
		md = self.ql.create_disassembler()

		self.ql.log.error('Disassembly:')

		for insn in tuple(md.disasm(data, address))[:num_insns]:
			opcodes = ''.join(f'{ch:02x}' for ch in insn.bytes[:10])

			if len(insn.bytes) > 10:
				opcodes += '.'

			self.ql.log.error(f'{insn.address:08x} :  {opcodes:<20s}  {insn.mnemonic:<10s} {insn.op_str:s}')

		self.ql.log.error(f'')


	def emu_error(self):
		dump_len = 64

		try:
			pc = self.ql.reg.arch_pc
			data = self.ql.mem.read(pc, dump_len)

			self.emit_context()
			self.emit_hexdump(pc, data)
			self.emit_disasm(pc, data)

			containing_image = self.find_containing_image(pc)
			img_info = f' ({containing_image.path} + {pc - containing_image.base:#x})' if containing_image else ''
			self.ql.log.error(f'PC = {pc:#010x}{img_info}')

			self.ql.log.error(f'Memory map:')
			self.ql.mem.show_mapinfo()
		except UcError:
			self.ql.log.error(f'Error: PC({pc:#x}) is unreachable')


	def run(self):
		self.notify_before_module_execution(self.ql, self.running_module)

		if self.ql.entry_point is not None:
			self.ql.loader.entry_point = self.ql.entry_point

		if self.ql.exit_point is not None:
			self.exit_point = self.ql.exit_point

		try:
			self.ql.emu_start(self.ql.loader.entry_point, self.exit_point, self.ql.timeout, self.ql.count)
		except KeyboardInterrupt as ex:
			self.ql.log.critical(f'Execution interrupted by user')

			if self.ql._internal_exception is ex:
				self.ql._internal_exception = None
		except UcError:
			self.emu_error()
			raise

		if self.ql._internal_exception is not None:
			raise self.ql._internal_exception



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *

def ql_syscall_uname(ql, address, *args, **kw):
    buf =  b''
    buf += b'QilingOS'.ljust(65, b'\x00')
    buf += b'ql_vm'.ljust(65, b'\x00')
    buf += b'99.0-RELEASE'.ljust(65, b'\x00')
    buf += b'QiligOS 99.0-RELEASE r1'.ljust(65, b'\x00')
    buf += b'ql_processor'.ljust(65, b'\x00')
    buf += b''.ljust(65, b'\x00')
    ql.mem.write(address, buf)
    regreturn = 0
    return regreturn



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from qiling.os.windows.const import *
from qiling.os.windows.fncc import *
from qiling.os.const import *
from qiling.os.windows.utils import *
from qiling.os.windows.thread import *
from qiling.os.windows.handle import *
from qiling.exception import *

dllname = 'cng_dll'

# typedef struct _OSVERSIONINFOW {
#   ULONG dwOSVersionInfoSize;
#   ULONG dwMajorVersion;
#   ULONG dwMinorVersion;
#   ULONG dwBuildNumber;
#   ULONG dwPlatformId;
#   WCHAR szCSDVersion[128];
# }
# NTSYSAPI NTSTATUS RtlGetVersion(
#   PRTL_OSVERSIONINFOW lpVersionInformation
# );
@winsdkapi(cc=CDECL, dllname=dllname, replace_params={"lpVersionInformation": POINTER})
def hook_RtlGetVersion(ql, address, params):
    pointer = params["lpVersionInformation"]
    size = int.from_bytes(ql.mem.read(pointer, 4), byteorder="little")
    os_version_info_asked = {
        "dwOSVersionInfoSize":
        size,
        VER_MAJORVERSION:
        int.from_bytes(ql.mem.read(pointer + 4, 4), byteorder="little"),
        VER_MINORVERSION:
        int.from_bytes(ql.mem.read(pointer + 8, 4), byteorder="little"),
        VER_BUILDNUMBER:
        int.from_bytes(ql.mem.read(pointer + 12, 4), byteorder="little"),
        VER_PLATFORMID:
        int.from_bytes(ql.mem.read(pointer + 16, 4), byteorder="little"),
        "szCSDVersion":
        int.from_bytes(ql.mem.read(pointer + 20, 128), byteorder="little"),
    }
    ql.mem.write(
        pointer + 4,
        ql.os.profile.getint("SYSTEM",
                             "majorVersion").to_bytes(4, byteorder="little"))
    ql.mem.write(
        pointer + 8,
        ql.os.profile.getint("SYSTEM",
                             "minorVersion").to_bytes(4, byteorder="little"))

    ql.log.debug("The sample is checking the windows Version!")
    return STATUS_SUCCESS



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#



from .const import *
from qiling.const import *

"""
set_tls
"""
def ql_arm_init_get_tls(ql):
    ql.mem.map(0xFFFF0000, 0x1000, info="[arm_tls]")
    """
    'adr r0, data; ldr r0, [r0]; mov pc, lr; data:.ascii "\x00\x00"'
    """
    sc = b'\x04\x00\x8f\xe2\x00\x00\x90\xe5\x0e\xf0\xa0\xe1\x00\x00\x00\x00'

    # if ql.archendian == QL_ENDIAN.EB:
    #    sc = ql.os.lsbmsb_convert(ql, sc)

    ql.mem.write(ql.arch.arm_get_tls_addr, sc)
    ql.log.debug("Set init_kernel_get_tls")    
         


#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import types

from qiling.os.linux.kernel_api import *
from qiling.exception import *


# hook Linux kernel API
def hook_kernel_api(ql, address, size):
    # call kernel api
    # print("check hook_kernel_api: %x" %(address))
    if address in ql.loader.import_symbols:
        api_name = ql.loader.import_symbols[address]
        # print("OK, found hook for %s" %api_name)
        api_func = None

        if api_name in ql.os.user_defined_api:
            if isinstance(ql.os.user_defined_api[api_name], types.FunctionType):
                api_func = ql.os.user_defined_api[api_name]
        else:
            try:
                api_func = globals()['hook_' + api_name]
            except KeyError:
                # print("ERROR on globals", globals())
                api_func = None

        if api_func:
            try:
                api_func(ql, address, {})
            except Exception:
                ql.log.exception("")
                ql.log.debug("%s Exception Found" % api_name)
                raise QlErrorSyscallError("Linux kernel API Implementation Error")
        else:
            ql.log.warning("%s is not implemented\n" % api_name)
            if ql.debug_stop:
                raise QlErrorSyscallNotFound("Linux kernel API Implementation Not Found")




#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from unicorn import *
from unicorn.arm64_const import *

from qiling.const import *
from .arch import QlArch
from .arm64_const import *

class QlArchARM64(QlArch):
    def __init__(self, ql):
        super(QlArchARM64, self).__init__(ql)

        register_mappings = [
            reg_map, reg_map_w
        ]

        for reg_maper in register_mappings:
            self.ql.reg.expand_mapping(reg_maper)            

        self.ql.reg.create_reverse_mapping()

        self.ql.reg.register_sp(reg_map["sp"])
        self.ql.reg.register_pc(reg_map["pc"])


    def stack_push(self, value):
        self.ql.reg.sp -= 8
        self.ql.mem.write(self.ql.reg.sp, self.ql.pack64(value))
        return self.ql.reg.sp


    def stack_pop(self):
        data = self.ql.unpack64(self.ql.mem.read(self.ql.reg.sp, 8))
        self.ql.reg.sp += 8
        return data


    def stack_read(self, offset):
        return self.ql.unpack64(self.ql.mem.read(self.ql.reg.sp + offset, 8))


    def stack_write(self, offset, data):
        return self.ql.mem.write(self.ql.reg.sp + offset, self.ql.pack64(data))


    # get initialized unicorn engine
    def get_init_uc(self):
        uc = Uc(UC_ARCH_ARM64, UC_MODE_ARM)   
        return uc


    def enable_vfp(self):
        self.ql.reg.cpacr_el1 = self.ql.reg.cpacr_el1 | 0x300000



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from inspect import signature

from unicorn.arm64_const import *
from unicorn.arm_const import *
from unicorn.mips_const import *
from unicorn.x86_const import *

from qiling.const import *
from qiling.os.os import QlOs
from qiling.utils import *
from qiling.exception import *

from qiling.os.posix.syscall import *
from qiling.os.linux.syscall import *
from qiling.os.macos.syscall import *
from qiling.os.freebsd.syscall import *

from qiling.os.linux.function_hook import ARMFunctionArg, MIPS32FunctionArg, ARM64FunctionArg, X86FunctionArg, X64FunctionArg


def getNameFromErrorCode(ret):
    """
    Return the hex representation of a return value and if possible
    add the corresponding error name to it.
    :param ret: Return value of a syscall.
    :return: The string representation of the error.
    """
    if -ret in errors:
        return hex(ret) + "(" + errors[-ret] + ")"
    else:
        return hex(ret)


class QlOsPosix(QlOs):
    def __init__(self, ql):
        super(QlOsPosix, self).__init__(ql)
        self.ql = ql
        self.sigaction_act = []
        
        if self.ql.root == True:
            self.uid = 0
            self.gid = 0
        else:    
            self.uid = self.profile.getint("KERNEL","uid")
            self.gid = self.profile.getint("KERNEL","gid")

        self.ipv6 = self.profile.getboolean("NETWORK","ipv6")
        self.bindtolocalhost = self.profile.getboolean("NETWORK","bindtolocalhost")
        self.fd = QlFileDes([0] * 256)
        self.dict_posix_syscall = dict()
        self.dict_posix_onEnter_syscall = dict()
        self.dict_posix_onExit_syscall = dict()
        self.dict_posix_syscall_by_num = dict()
        self.dict_posix_onEnter_syscall_by_num = dict()
        self.dict_posix_onExit_syscall_by_num = dict()

        self.syscall_map = None
        self.syscall_name = None

        if self.ql.ostype in QL_OS_POSIX:
            self.fd[0] = self.stdin
            self.fd[1] = self.stdout
            self.fd[2] = self.stderr

        for _ in range(256):
            self.sigaction_act.append(0)

    # ql.syscall - get syscall for all posix series
    @property
    def syscall(self):
        return self.get_syscall()

    # ql.func_arg - get syscall for all posix series
    @property
    def function_arg(self):
        if self.ql.ostype in (QL_OS_POSIX):
            # ARM
            if self.ql.archtype== QL_ARCH.ARM:
                return ARMFunctionArg(self.ql)

            # MIPS32
            elif self.ql.archtype== QL_ARCH.MIPS:
                return MIPS32FunctionArg(self.ql)

            # ARM64
            elif self.ql.archtype== QL_ARCH.ARM64:
                return ARM64FunctionArg(self.ql)

            # X86
            elif  self.ql.archtype== QL_ARCH.X86:
                return X86FunctionArg(self.ql)

            # X8664
            elif  self.ql.archtype== QL_ARCH.X8664:
                return X64FunctionArg(self.ql)
            else:
                raise

    def load_syscall(self, intno=None):
        # import syscall mapping function
        map_syscall = ql_syscall_mapping_function(self.ql.ostype)
        self.syscall_name = map_syscall(self.ql, self.syscall)

        if self.dict_posix_onEnter_syscall.get(self.syscall_name) != None:
            self.syscall_onEnter = self.dict_posix_onEnter_syscall.get(self.syscall_name)
        elif self.dict_posix_onEnter_syscall_by_num.get(self.syscall) != None:
            self.syscall_onEnter = self.dict_posix_onEnter_syscall_by_num.get(self.syscall)
        else:
            self.syscall_onEnter = None    
        
        if self.dict_posix_onExit_syscall.get(self.syscall_name) != None:
            self.syscall_onExit = self.dict_posix_onExit_syscall.get(self.syscall_name)
        elif self.dict_posix_onExit_syscall_by_num.get(self.syscall) != None:
            self.syscall_onExit = self.dict_posix_onExit_syscall_by_num.get(self.syscall)
        else:
            self.syscall_onExit = None    
        
        self.syscall_map = self.dict_posix_syscall_by_num.get(self.syscall)
        syscall_name_str = None
        

        if self.syscall_map is not None:
            self.syscall_name = self.syscall_map.__name__
        else:
            self.syscall_name = map_syscall(self.ql, self.syscall)
            _ostype_str = ostype_convert_str(self.ql.ostype)
            _posix_syscall = ql_get_module_function(f"qiling.os.posix", "syscall")
            _os_syscall = ql_get_module_function(f"qiling.os.{_ostype_str.lower()}", "syscall")
            
            if self.syscall_name not in dir(_posix_syscall) \
            and self.syscall_name not in dir(_os_syscall):

                syscall_name_str = self.syscall_name
                self.syscall_map = None
                self.syscall_name = None

                
            if self.syscall_name is not None:
                replace_func = self.dict_posix_syscall.get(self.syscall_name)
                if replace_func is not None:
                    self.syscall_map = replace_func
                    self.syscall_name = replace_func.__name__
                else:
                    self.syscall_map = eval(self.syscall_name)
            else:
                self.syscall_map = None
                self.syscall_name = None

        if self.syscall_map is not None:
            self.syscalls.setdefault(self.syscall_name, []).append({
                "params": {
                    "param0": self.get_func_arg()[0],
                    "param1": self.get_func_arg()[1],
                    "param2": self.get_func_arg()[2],
                    "param3": self.get_func_arg()[3],
                    "param4": self.get_func_arg()[4],
                    "param5": self.get_func_arg()[5]
                },
                "result": None,
                "address": self.ql.reg.arch_pc,
                "return_address": None,
                "position": self.syscalls_counter
            })

            self.syscalls_counter += 1

            try:                
                if self.syscall_onEnter is None:
                    ret = 0
                else:
                    ret = self.syscall_onEnter(self.ql, self.get_func_arg()[0], self.get_func_arg()[1], self.get_func_arg()[2], self.get_func_arg()[3], self.get_func_arg()[4], self.get_func_arg()[5])

                if isinstance(ret, int) == False or ret & QL_CALL_BLOCK == 0:
                    args = []
                    for n, argname in enumerate(signature(self.syscall_map).parameters.values()):
                        argname = str(argname)
                        if not n or argname.startswith("*"):
                            # first arg for syscalls is ql
                            continue
                        else:
                            # cut the first part of the arg if it is of form fstatat64_fd
                            argname = argname if "_" not in argname else "".join(argname.split("_")[1:])
                            args.append(f"{argname}={hex(self.get_func_arg()[n-1])}")
                    args = ", ".join(args)
                    self.ql.log.info("0x%x: %s(%s)" % (self.ql.reg.arch_pc, self.syscall_map.__name__[11:], args))
                    ret = self.syscall_map(self.ql, self.get_func_arg()[0], self.get_func_arg()[1], self.get_func_arg()[2], self.get_func_arg()[3], self.get_func_arg()[4], self.get_func_arg()[5])
                    if ret is not None and isinstance(ret, int):
                        # each name has a list of calls, we want the last one and we want to update the return value
                        self.syscalls[self.syscall_name][-1]["result"] = ret
                        ret = self.set_syscall_return(ret)
                        self.ql.log.debug("%s() = %s" % (self.syscall_map.__name__[11:], getNameFromErrorCode(ret)))

                if self.syscall_onExit is not None:
                    self.syscall_onExit(self.ql, self.get_func_arg()[0], self.get_func_arg()[1], self.get_func_arg()[2], self.get_func_arg()[3], self.get_func_arg()[4], self.get_func_arg()[5])

            except KeyboardInterrupt:
                raise
            except Exception as e:
                self.ql.log.exception("")
                self.ql.log.info("Syscall ERROR: %s DEBUG: %s" % (self.syscall_name, e))
                raise e
        else:
            self.ql.log.warning(
                "0x%x: syscall %s number = 0x%x(%d) not implemented" % (self.ql.reg.arch_pc, syscall_name_str, self.syscall, self.syscall))
            if self.ql.debug_stop:
                raise QlErrorSyscallNotFound("Syscall Not Found")

    # get syscall
    def get_syscall(self):
        if self.ql.archtype == QL_ARCH.ARM64:
            if self.ql.ostype == QL_OS.MACOS:
                syscall_num = UC_ARM64_REG_X16
            else:
                syscall_num = UC_ARM64_REG_X8
        elif self.ql.archtype == QL_ARCH.ARM:
            syscall_num = UC_ARM_REG_R7
        elif self.ql.archtype == QL_ARCH.MIPS:
            syscall_num = UC_MIPS_REG_V0
        elif self.ql.archtype == QL_ARCH.X86:
            syscall_num = UC_X86_REG_EAX
        elif self.ql.archtype == QL_ARCH.X8664:
            syscall_num = UC_X86_REG_RAX

        return self.ql.reg.read(syscall_num)

    def set_syscall_return(self, regreturn):
        if self.ql.archtype == QL_ARCH.ARM:  # ARM
            self.ql.reg.r0 = regreturn

        elif self.ql.archtype == QL_ARCH.ARM64:  # ARM64
            self.ql.reg.x0 = regreturn

        elif self.ql.archtype == QL_ARCH.X86:  # X86
            self.ql.reg.eax = regreturn

        elif self.ql.archtype == QL_ARCH.X8664:  # X8664
            self.ql.reg.rax = regreturn

        elif self.ql.archtype == QL_ARCH.MIPS:  # MIPSE32EL
            if regreturn < 0 and regreturn > -1134:
                a3return = 1
                regreturn = - regreturn
            else:
                a3return = 0

            self.ql.reg.v0 = regreturn
            self.ql.reg.a3 = a3return
        return regreturn

    # get syscall
    def get_func_arg(self):
        if self.ql.archtype == QL_ARCH.ARM64:
            param0 = self.ql.reg.x0
            param1 = self.ql.reg.x1
            param2 = self.ql.reg.x2
            param3 = self.ql.reg.x3
            param4 = self.ql.reg.x4
            param5 = self.ql.reg.x5
        elif self.ql.archtype == QL_ARCH.ARM:
            param0 = self.ql.reg.r0
            param1 = self.ql.reg.r1
            param2 = self.ql.reg.r2
            param3 = self.ql.reg.r3
            param4 = self.ql.reg.r4
            param5 = self.ql.reg.r5
        elif self.ql.archtype == QL_ARCH.MIPS:
            param0 = self.ql.reg.a0
            param1 = self.ql.reg.a1
            param2 = self.ql.reg.a2
            param3 = self.ql.reg.a3
            param4 = self.ql.reg.sp
            param4 = param4 + 0x10
            param5 = self.ql.reg.sp
            param5 = param5 + 0x14
        elif self.ql.archtype == QL_ARCH.X86:
            param0 = self.ql.reg.ebx
            param1 = self.ql.reg.ecx
            param2 = self.ql.reg.edx
            param3 = self.ql.reg.esi
            param4 = self.ql.reg.edi
            param5 = self.ql.reg.ebp
        elif self.ql.archtype == QL_ARCH.X8664:
            param0 = self.ql.reg.rdi
            param1 = self.ql.reg.rsi
            param2 = self.ql.reg.rdx
            param3 = self.ql.reg.r10
            param4 = self.ql.reg.r8
            param5 = self.ql.reg.r9

        return [param0, param1, param2, param3, param4, param5]



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import ntpath, os, uuid

from sys import getsizeof

from qiling.const import *

from qiling.os.const import *
from .registry import RegistryManager
from .clipboard import Clipboard
from .fiber import FiberManager
from .handle import HandleManager, Handle
from .thread import QlWindowsThreadManagement, QlWindowsThread
from .structs import UNICODE_STRING32, UNICODE_STRING64


def ql_x86_windows_hook_mem_error(ql, access, addr, size, value):
    ql.log.debug("ERROR: unmapped memory access at 0x%x" % addr)
    return False


def string_unpack(string):
    return string.decode().split("\x00")[0]


def read_guid(ql, address):
    result = ""
    raw_guid = ql.mem.read(address, 16)
    return uuid.UUID(bytes_le=bytes(raw_guid))


def env_dict_to_array(env_dict):
    env_list = []
    for item in env_dict:
        env_list.append(item + "=" + env_dict[item])
    return env_list


def debug_print_stack(ql, num, message=None):
    if message:
        ql.log.debug("========== %s ==========" % message)
        sp = ql.reg.arch_sp
        ql.log.debug(hex(sp + ql.pointersize * num) + ": " + hex(ql.stack_read(num * ql.pointersize)))


def is_file_library(string):
    string = string.lower()
    extension = string[-4:]
    return extension in (".dll", ".exe", ".sys", ".drv")


def string_to_hex(string):
    return ":".join("{:02x}".format(ord(c)) for c in string)


def string_appearance(ql, string):
    strings = string.split(" ")
    for string in strings:
        val = ql.os.appeared_strings.get(string, set())
        val.add(ql.os.syscalls_counter)
        ql.os.appeared_strings[string] = val


def path_leaf(path):
    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)


def find_size_function(ql, func_addr):
    # We have to retrieve the return address position
    code = ql.mem.read(func_addr, 0x100)
    return_procedures = [b"\xc3", b"\xc2", b"\xcb", b"\xca"]
    min_index = min([code.index(return_value) for return_value in return_procedures if return_value in code])
    return min_index



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import struct


from qiling.os.windows.const import *
from qiling.os.windows.fncc import *
from qiling.os.const import *
from qiling.os.windows.utils import *
from qiling.os.windows.thread import *
from qiling.os.windows.handle import *
from qiling.exception import *
import qiling.os.windows.structs

dllname = 'ntdll_dll'

# void *memcpy(
#    void *dest,
#    const void *src,
#    size_t count
# );
@winsdkapi(cc=CDECL, dllname=dllname, replace_params={"dest": POINTER, "src": POINTER, "count": UINT})
def hook_memcpy(ql, address, params):
    try:
        data = bytes(ql.mem.read(params['src'], params['count']))
        ql.mem.write(params['dest'], data)
    except Exception as e:
        ql.log.exception("")
    return params['dest']


def _QueryInformationProcess(ql, address, params):
    flag = params["ProcessInformationClass"]
    dst = params["ProcessInformation"]
    pt_res = params["ReturnLength"]
    if flag == ProcessDebugFlags:
        value = b"\x01" * 0x4
    elif flag == ProcessDebugPort:
        value = b"\x00" * 0x4
    elif flag == ProcessDebugObjectHandle:
        return STATUS_PORT_NOT_SET
    elif flag == ProcessBasicInformation:
        pbi = qiling.os.windows.structs.ProcessBasicInformation(ql, exitStatus=0,
                                                                pebBaseAddress=ql.os.heap_base_address, affinityMask=0,
                                                                basePriority=0,
                                                                uniqueId=ql.os.profile.getint("KERNEL", "pid"),
                                                                parentPid=ql.os.profile.getint("KERNEL", "parent_pid"))
        addr = ql.os.heap.alloc(pbi.size)
        pbi.write(addr)
        value = addr.to_bytes(ql.pointersize, "little")
    else:
        ql.log.debug(str(flag))
        raise QlErrorNotImplemented("API not implemented")
    ql.log.debug("The target is checking the debugger via QueryInformationProcess ")
    ql.mem.write(dst, value)
    if pt_res != 0:
        ql.mem.write(pt_res, 0x8.to_bytes(1, byteorder="little"))

    return STATUS_SUCCESS

# NTSTATUS WINAPI ZwQueryInformationProcess(
#   _In_      HANDLE           ProcessHandle,
#   _In_      PROCESSINFOCLASS ProcessInformationClass,
#   _Out_     PVOID            ProcessInformation,
#   _In_      ULONG            ProcessInformationLength,
#   _Out_opt_ PULONG           ReturnLength
# );
@winsdkapi(cc=CDECL, dllname=dllname,
           replace_params={"ProcessHandle": HANDLE, "ProcessInformationClass": INT, "ProcessInformation": POINTER,
                      "ProcessInformationLength": UINT, "ReturnLength": POINTER})
def hook_ZwQueryInformationProcess(ql, address, params):
    # TODO have no idea if is cdecl or stdcall

    _QueryInformationProcess(ql, address, params)


# __kernel_entry NTSTATUS NtQueryInformationProcess(
#   IN HANDLE           ProcessHandle,
#   IN PROCESSINFOCLASS ProcessInformationClass,
#   OUT PVOID           ProcessInformation,
#   IN ULONG            ProcessInformationLength,
#   OUT PULONG          ReturnLength
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_NtQueryInformationProcess(ql, address, params):
    # TODO have no idea if is cdecl or stdcall

    _QueryInformationProcess(ql, address, params)

def _QuerySystemInformation(ql, address, params):
    siClass = params["SystemInformationClass"]
    pt_res = params["ReturnLength"]
    dst = params["SystemInformation"]
    if (siClass == SystemBasicInformation):
        bufferLength = params["SystemInformationLength"]
        if (ql.archtype == QL_ARCH.X8664):
            sbi = qiling.os.windows.structs.SystemBasicInforation(ql,
                                                              Reserved=0,
                                                              TimerResolution = 156250 ,
                                                              PageSize=ql.os.heap.page_size,
                                                              NumberOfPhysicalPages = 0x003FC38A,
                                                              LowestPhysicalPageNumber=1,
                                                              HighestPhysicalPageNumber=0x0046DFFF,
                                                              AllocationGranularity=1,
                                                              MinimumUserModeAddress=0x10000,
                                                              MaximumUserModeAddress=0x7FFFFFFEFFFF,
                                                              ActiveProcessorsAffinityMask = 0x3F,
                                                              NumberOfProcessors = 0x6)
        elif ql.archtype == QL_ARCH.X86:
            sbi = qiling.os.windows.structs.SystemBasicInforation(ql,
                                                                  Reserved=0,
                                                                  TimerResolution=156250,
                                                                  PageSize=ql.os.heap.page_size,
                                                                  NumberOfPhysicalPages=0x003FC38A,
                                                                  LowestPhysicalPageNumber=1,
                                                                  HighestPhysicalPageNumber=0x0046DFFF,
                                                                  AllocationGranularity=1,
                                                                  MinimumUserModeAddress=0x10000,
                                                                  MaximumUserModeAddress=0x7FFEFFFF,
                                                                  ActiveProcessorsAffinityMask=0x3F,
                                                                  NumberOfProcessors=0x6)
        if (bufferLength==sbi.size):
            sbi.write(dst)
            if pt_res != 0:
                ql.mem.write(pt_res, sbi.size.to_bytes(1, byteorder="little"))
        else:
            if pt_res != 0:
                ql.mem.write(pt_res, sbi.size.to_bytes(1, byteorder="little"))
            return STATUS_INFO_LENGTH_MISMATCH
    else:
        ql.log.debug(str(siClass))
        raise QlErrorNotImplemented("API not implemented")


    return STATUS_SUCCESS

# __kernel_entry NTSTATUS NtQuerySystemInformation(
#   IN SYSTEM_INFORMATION_CLASS SystemInformationClass,
#   OUT PVOID                   SystemInformation,
#   IN ULONG                    SystemInformationLength,
#   OUT PULONG                  ReturnLength
# );
@winsdkapi(cc=STDCALL, dllname=dllname,
                replace_params={"SystemInformationClass": UINT, "SystemInformation": POINTER, "SystemInformationLength": SIZE_T,
                    "ReturnLength": POINTER})
def hook_NtQuerySystemInformation(ql, address, params):
    # In minwindef.h
    # #define WINAPI      __stdcall

    _QuerySystemInformation(ql, address, params)

# pub unsafe extern "system" fn ZwQuerySystemInformation(
#   IN SYSTEM_INFORMATION_CLASS SystemInformationClass,
#   OUT PVOID                   SystemInformation,
#   IN ULONG                    SystemInformationLength,
#   OUT PULONG                  ReturnLength
# );
@winsdkapi(cc=STDCALL, dllname=dllname,
                replace_params={"SystemInformationClass": UINT, "SystemInformation": POINTER, "SystemInformationLength": SIZE_T,
                    "ReturnLength": POINTER})
def hook_ZwQuerySystemInformation(ql, address, params):
    # In minwindef.h
    # #define WINAPI      __stdcall

    return _QuerySystemInformation(ql, address, params)

# pub unsafe extern "system" fn ZwCreateDebugObject(
#     DebugObjectHandle: PHANDLE,
#     DesiredAccess: ACCESS_MASK,
#     ObjectAttributes: POBJECT_ATTRIBUTES,
#     Flags: ULONG
# ) -> NTSTATUS
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={"DebugObjectHandle": HANDLE, "DesiredAccess": INT,
                                                   "ObjectAttributes": POINTER, "Flags": ULONGLONG})
def hook_ZwCreateDebugObject(ql, address, params):
    # FIXME: find documentation, almost none was found online, and create the correct object
    handle = Handle(id=params["DebugObjectHandle"])
    ql.os.handle_manager.append(handle)
    return STATUS_SUCCESS


# __kernel_entry NTSYSCALLAPI NTSTATUS NtQueryObject(
#   HANDLE                   Handle,
#   OBJECT_INFORMATION_CLASS ObjectInformationClass,
#   PVOID                    ObjectInformation,
#   ULONG                    ObjectInformationLength,
#   PULONG                   ReturnLength
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_ZwQueryObject(ql, address, params):
    infoClass = params["ObjectInformationClass"]
    dest = params["ObjectInformation"]
    size_dest = params["ReturnLength"]
    string = "DebugObject".encode("utf-16le")
    string_addr = ql.os.heap.alloc(len(string))
    ql.log.debug(str(string_addr))
    ql.log.debug(str(string))
    ql.mem.write(string_addr, string)
    us = qiling.os.windows.structs.UnicodeString(ql, length=len(string), maxLength=len(string),
                                                 buffer=string_addr)

    if infoClass == ObjectTypeInformation:
        res = qiling.os.windows.structs.ObjectTypeInformation(ql, us, 1, 1)
    elif infoClass == ObjectAllTypesInformation:
        # FIXME: there is an error in how these structs are read by al-khaser. Have no idea on where, so we are
        #  bypassing it
        # oti = qiling.os.windows.structs.ObjectTypeInformation(ql, us, 1, 1)
        # res = qiling.os.windows.structs.ObjectAllTypesInformation(ql, 2, oti)
        return 1
    else:
        raise QlErrorNotImplemented("API not implemented")
    if dest != 0 and params["Handle"] != 0:
        res.write(dest)
    if size_dest != 0:
        ql.mem.write(size_dest, res.size.to_bytes(4, "little"))

    return STATUS_SUCCESS


# NTSYSAPI NTSTATUS NTAPI NtSetInformatonProcess(
#   _In_      HANDLE           ProcessHandle,
#   _In_      PROCESSINFOCLASS ProcessInformationClass,
#   _In_      PVOID            ProcessInformation
#   _In_      ULONG            ProcessInformationLength
# );
@winsdkapi(cc=STDCALL, dllname=dllname,
            replace_params={"ProcessHandle": HANDLE, "ProcessInformationClass": INT, "ProcessInformation": POINTER,
                        "ProcessInformationLength": UINT})
def hook_ZwSetInformationProcess(ql, address, params):
    _SetInformationProcess(ql, address, params)

def hook_NtSetInformationProcess(ql, address, params):
    _SetInformationProcess(ql, address, params)

def _SetInformationProcess(ql, address, params):
    process = params["ProcessHandle"]
    flag = params["ProcessInformationClass"]
    dst = params["ProcessInformation"]
    pt_res = params["ReturnLength"]
    
    if flag == ProcessDebugFlags:
        value = b"\x01" * 0x4
    elif flag == ProcessDebugPort:
        value = b"\x00" * 0x4
    elif flag == ProcessDebugObjectHandle:
        return STATUS_PORT_NOT_SET
    elif flag == ProcessBreakOnTermination:
            ql.log.debug("The target may be attempting modify a the 'critical' flag of the process")  
    elif flag  == ProcessExecuteFlags:
        ql.log.debug("The target may be attempting to modify DEP for the process")
        if dst != 0:
            ql.mem.write(dst, 0x0.to_bytes(1, byteorder="little"))

    elif flag == ProcessBasicInformation:
        pbi = qiling.os.windows.structs.ProcessBasicInformation(
            ql, exitStatus=0,
            pebBaseAddress=ql.os.heap_base_address, affinityMask=0,
            basePriority=0,
            uniqueId=ql.os.profile.getint("KERNEL", "pid"),
            parentPid=ql.os.profile.geting("KERNEL", "parent_pid")
        )
        ql.log.debug("The target may be attempting to modify the PEB debug flag")
        addr = ql.os.heap.alloc(pbi.size)
        pbi.write(addr)
        value = addr.to_bytes(ql.pointersize, "little")
    else:
        ql.log.debug(str(flag))
        raise QlErrorNotImplemented("API not implemented")

    return STATUS_SUCCESS

# NTSYSAPI
# NTSTATUS
# NTAPI
# NtYieldExecution(
#  );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_ZwYieldExecution(ql, address, params):
    # FIXME: offer timeslice of this thread
    return STATUS_NO_YIELD_PERFORMED


# NTSTATUS LdrGetProcedureAddress(
#  IN HMODULE              ModuleHandle,
#  IN PANSI_STRING         FunctionName OPTIONAL,
#  IN WORD                 Oridinal OPTIONAL,
#  OUT PVOID               *FunctionAddress );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={"ModuleHandle": POINTER,
    "FunctionName": STRING, "Ordinal": UINT, "FunctionAddress": POINTER})
def hook_LdrGetProcedureAddress(ql, address, params):
    if params['FunctionName']:
        identifier = bytes(params["lpProcName"], 'ascii')
    else:
        identifier = params['Ordinal']
    # Check if dll is loaded
    try:
        dll_name = [key for key, value in ql.loader.dlls.items() if value == params['ModuleHandle']][0]
    except IndexError as ie:
        ql.log.info('Failed to import function "%s" with handle 0x%X' % (lpProcName, params['ModuleHandle']))
        return 0

    if identifier in ql.loader.import_address_table[dll_name]:
        addr = ql.loader.import_address_table[dll_name][identifier]
        ql.mem.write(addr.to_bytes(length=ql.pointersize, byteorder='little'), params['FunctionAddress'])
        return 0

    return 0xFFFFFFFF


# NTSYSAPI PVOID RtlAllocateHeap(
#  PVOID  HeapHandle,
#  ULONG  Flags,
#  SIZE_T Size
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={"HeapHandle": POINTER,
    "Flags": UINT,"Size": SIZE_T})
def hook_RtlAllocateHeap(ql, address, params):
    ret = ql.os.heap.alloc(params["Size"])
    return ret


# wchar_t* wcsstr( const wchar_t* dest, const wchar_t* src );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={"dest": POINTER, "src": WSTRING})
def hook_wcsstr(ql, address, params):
    dest = params["dest"]
    value = ql.os.read_wstring(dest)
    params["dest"] = value
    src = params["src"]
    if src in value:
        pos = value.index(src)
        return dest + post
    return 0


# HANDLE CsrGetProcessId();
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_CsrGetProcessId(ql, address, params):
    pid = ql.os.profile["PROCESSES"].getint("csrss.exe", fallback=12345)
    return pid


#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import struct

from qiling.exception import *
from qiling.const import *
from qiling.arch.x86_const import *
from qiling.os.posix.const_mapping import *
from qiling.os.filestruct import *

from .const import *
from .thread import *
from .mach_port import *
from .kernel_func import *
from .utils import *

# TODO: We need to finish these syscall
# there are three kinds of syscall, we often use posix syscall, mach syscall is used by handle mach msg
# Unfortunately we dont have enough doc about mach syscallios 
# We can find all of these syscalls in kernel source code, some pthread func may found in libpthread

################
# ios syscall #
################

def ql_syscall_fgetattrlist(ql, fd, alist, attributeBuffer, bufferSize, options, *args, **kw):
    ql.log.debug("fgetattrlist(fd: 0x%x, alist: 0x%x, attributeBuffer: 0x%x, bufferSize: 0x%x, options: 0x%x)" % (
        fd, alist, attributeBuffer, bufferSize, options
    ))

    attrlist = {}
    attrlist["bitmapcount"] = unpack("<H", ql.mem.read(alist, 2))[0]
    attrlist["reserved"] = unpack("<H", ql.mem.read(alist + 2, 2))[0]
    attrlist["commonattr"] = unpack("<L", ql.mem.read(alist + 4, 4))[0]
    attrlist["volattr"] = unpack("<L", ql.mem.read(alist + 8, 4))[0]
    attrlist["dirattr"] = unpack("<L", ql.mem.read(alist + 12, 4))[0]
    attrlist["fileattr"] = unpack("<L", ql.mem.read(alist + 16, 4))[0]
    attrlist["forkattr"] = unpack("<L", ql.mem.read(alist + 20, 4))[0]

    ql.log.debug("bitmapcount: 0x%x, reserved: 0x%x, commonattr: 0x%x, volattr: 0x%x, dirattr: 0x%x, fileattr: 0x%x, forkattr: 0x%x\n" % (
        attrlist["bitmapcount"], attrlist["reserved"], attrlist["commonattr"], attrlist["volattr"], attrlist["dirattr"], attrlist["fileattr"], attrlist["forkattr"]
    ))

    # path_str = macho_read_string(ql, path, MAX_PATH_SIZE)

    attr = b''
    if attrlist["commonattr"] != 0:
        commonattr = ql.os.macho_fs.get_common_attr(ql.path, attrlist["commonattr"])
        if not commonattr:
            raise QlErrorSyscallError("Error File Not Exist")
        attr += commonattr
    
    attr_len = len(attr) + 4
    attr = struct.pack("<L", attr_len) + attr

    if len(attr) > bufferSize:
        ql.log.debug("Length error")
        return 1
    else:

        ql.mem.write(attributeBuffer, attr)
        #set_eflags_cf(ql, 0x0)
        return KERN_SUCCESS


def ql_syscall_poll(ql, target, address, size, *args, **kw):
    return KERN_SUCCESS


################
# mach syscall #
################

# 0xa
def ql_syscall_kernelrpc_mach_vm_allocate_trap(ql, port, addr, size, flags, *args, **kw):
    ql.log.debug("[mach] mach vm allocate trap(port: 0x%x, addr: 0x%x, size: 0x%x, flags: 0x%x" % (port, addr, size, flags))
    mmap_address = ql.os.macho_task.min_offset
    mmap_end = page_align_end(mmap_address + size, PAGE_SIZE)
    ql.mem.map(mmap_address, mmap_end - mmap_address)
    ql.mem.write(mmap_address, b'\x00'*(mmap_end - mmap_address))
    ql.os.macho_task.min_offset = mmap_end
    ql.log.debug("vm alloc form 0x%x to 0x%0x" % (mmap_address, mmap_end))
    ql.mem.write(addr, struct.pack("<Q", mmap_address))
    return 0

# 0xc
def ql_syscall_kernelrpc_mach_vm_deallocate_trap(ql, target, address, size, *args, **kw):
    ql.log.debug("[mach] mach vm deallocate trap")
    return KERN_SUCCESS

# 0xf
def ql_syscall_kernelrpc_mach_vm_map_trap(ql, target, address, size, mask, flags, cur_protection):
    ql.log.debug("[mach] mach vm map trap(target: 0x%x, address: 0x%x, size: 0x%x, mask: 0x%x, flag: 0x%x, cur_protect: 0x%x)" % (
        target, address, size, mask, flags, cur_protection
    ))

    if ql.os.macho_vmmap_end & mask > 0:
        ql.os.macho_vmmap_end = ql.os.macho_vmmap_end - (ql.os.macho_vmmap_end & mask)
        ql.os.macho_vmmap_end += mask + 1

    
    vmmap_address = page_align_end(ql.os.macho_vmmap_end, PAGE_SIZE)
    vmmap_end = page_align_end(vmmap_address + size, PAGE_SIZE)

    ql.os.macho_vmmap_end = vmmap_end
    ql.mem.map(vmmap_address, vmmap_end - vmmap_address)
    ql.mem.write(address, struct.pack("<Q", vmmap_address))
    return KERN_SUCCESS

# 0x12
def ql_syscall_kernelrpc_mach_port_deallocate_trap(ql, *args, **kw):
    ql.log.debug("[mach] mach port deallocate trap")

# 0x13
def ql_syscall_kernelrpc_mach_port_mod_refs_trap(ql, target, name, right, delta, *args, **kw):
    ql.log.debug("[mach] mach port mod refs trap(target: 0x%x, name: 0x%x, right: 0x%x, delta: 0x%x)" % (
        target, name, right, delta
    ))
    pass

# 0x18
def ql_syscall_kernelrpc_mach_port_construct_trap(ql, target, options, context, name, *args, **kw):
    ql.log.debug("[mach] mach port construct trap(target: 0x%x, options: 0x%x, context: 0x%x, name: 0x%x)" % (
        target, options, context, name
    ))
    pass

# 0x1a
def ql_syscall_mach_reply_port(ql, *args, **kw):
    ql.log.debug("[mach] mach reply port , ret: %s" % (ql.os.macho_mach_port.name))
    return ql.os.macho_mach_port.name

# 0x1b
def ql_syscall_thread_self_trap(ql, *args, **kw):
    port_manager = ql.os.macho_port_manager
    thread_port = port_manager.get_thread_port(ql.os.macho_thread)
    ql.log.debug("[mach] thread_self_trap: ret: %s" % (thread_port))
    return thread_port

# 0x1c
def ql_syscall_task_self_trap(ql, *args, **kw):
    ql.log.debug("[mach] task self trap, ret: %d" % (ql.os.macho_task.id))
    return ql.os.macho_task.id

# 0x1d
def ql_syscall_host_self_trap(ql, *args, **kw):
    port_manager = ql.os.macho_port_manager
    ql.log.debug("[mach] host_self_trap, ret: %s" % (ql.os.macho_port_manager.host_port.name))
    return port_manager.host_port.name

# 0x1f
def ql_syscall_mach_msg_trap(ql, args, opt, ssize, rsize, rname, timeout):
    ql.log.debug("[mach] mach_msg_trap(args: 0x%x opt: 0x%x, ssize: 0x%x, rsize: 0x%x, rname: 0x%x, timeout: %d)" % (
        args, opt, ssize, rsize, rname, timeout))
    mach_msg = MachMsg(ql)
    mach_msg.read_msg_from_mem(args, ssize)
    ql.log.debug("Recv-> Header: %s, Content: %s" % (mach_msg.header, mach_msg.content))
    ql.os.macho_port_manager.deal_with_msg(mach_msg, args)
    return 0


#################
# POSIX syscall #
#################

# 0x21
def ql_syscall_access_macos(ql, path, flags, *args, **kw):
    path_str = ql.os.read_cstring(path)
    ql.log.debug("access(path: %s, flags: 0x%x)" % (path_str, flags))
    if not ql.os.macho_fs.isexists(path_str):
        return ENOENT
    else:
        return KERN_SUCCESS

# 0x30 
def ql_syscall_sigprocmask(ql, how, mask, omask, *args, **kw):
    ql.log.debug("sigprocmask(how: 0x%x, mask: 0x%x, omask: 0x%x)" % (how, mask, omask))

# 0x5c
def ql_syscall_fcntl64_macos(ql, fcntl_fd, fcntl_cmd, fcntl_arg, *args, **kw):
    regreturn = 0
    if fcntl_cmd == F_GETFL:
        regreturn = 2
    elif fcntl_cmd == F_SETFL:
        regreturn = 0
    elif fcntl_cmd == F_GETFD:
        regreturn = 2
    elif fcntl_cmd == F_SETFD:
        regreturn = 0
    elif fcntl_cmd == F_ADDFILESIGS_RETURN:
        ql.mem.write(fcntl_arg, ql.pack32(0xefffffff))
        regreturn = 0
    else:
        regreturn = 0

    ql.log.debug("fcntl64(fd: %d, cmd: %d, arg: 0x%x) = %d" % (fcntl_fd, fcntl_cmd, fcntl_arg, regreturn))
    return regreturn

# 0x99
def ql_syscall_pread(ql, fd, buf, nbyte, offset, *args, **kw):
    ql.log.debug("pread(fd: 0x%x, buf: 0x%x, nbyte: 0x%x, offset: 0x%x)" % (
        fd, buf, nbyte, offset
    ))
    if fd >= 0 and fd <= MAX_FD_SIZE:
        ql.os.fd[fd].lseek(offset)
        data = ql.os.fd[fd].read(nbyte)
        ql.mem.write(buf, data)
    set_eflags_cf(ql, 0x0)
    return nbyte

# 0xa9
def ql_syscall_csops(ql, pid, ops, useraddr, usersize, *args, **kw):
    flag = struct.pack("<L", (CS_ENFORCEMENT | CS_GET_TASK_ALLOW))
    ql.mem.write(useraddr, flag)
    ql.log.debug("csops(pid: %d, ops: 0x%x, useraddr: 0x%x, usersize: 0x%x) flag: 0x%x" % (
        pid, ops, useraddr, usersize, ((CS_ENFORCEMENT | CS_GET_TASK_ALLOW))
    ))
    return KERN_SUCCESS

# 0xdc
def ql_syscall_getattrlist(ql, path, alist, attributeBuffer, bufferSize, options, *args, **kw):
    ql.log.debug("getattrlist(path: 0x%x, alist: 0x%x, attributeBuffer: 0x%x, bufferSize: 0x%x, options: 0x%x)" % (
        path, alist, attributeBuffer, bufferSize, options
    ))
    attrlist = {}
    attrlist["bitmapcount"] = unpack("<H", ql.mem.read(alist, 2))[0]
    attrlist["reserved"] = unpack("<H", ql.mem.read(alist + 2, 2))[0]
    attrlist["commonattr"] = unpack("<L", ql.mem.read(alist + 4, 4))[0]
    attrlist["volattr"] = unpack("<L", ql.mem.read(alist + 8, 4))[0]
    attrlist["dirattr"] = unpack("<L", ql.mem.read(alist + 12, 4))[0]
    attrlist["fileattr"] = unpack("<L", ql.mem.read(alist + 16, 4))[0]
    attrlist["forkattr"] = unpack("<L", ql.mem.read(alist + 20, 4))[0]
    path_str = ql.os.read_cstring(path)

    ql.log.debug("bitmapcount: 0x%x, reserved: 0x%x, commonattr: 0x%x, volattr: 0x%x, dirattr: 0x%x, fileattr: 0x%x, forkattr: 0x%x\n" % (
        attrlist["bitmapcount"], attrlist["reserved"], attrlist["commonattr"], attrlist["volattr"], attrlist["dirattr"], attrlist["fileattr"], attrlist["forkattr"]
    ))
    ql.log.debug("path str: %s\n" % (path_str))

    attr = b''
    if attrlist["commonattr"] != 0:
        commonattr = ql.os.macho_fs.get_common_attr(path_str, attrlist["commonattr"])
        if not commonattr:
            ql.log.debug("Error File Not Exist: %s" % (path_str))
            raise QlErrorSyscallError("Error File Not Exist")
        attr += commonattr
    
    attr_len = len(attr) + 4
    attr = struct.pack("<L", attr_len) + attr

    if len(attr) > bufferSize:
        ql.log.debug("Length error")
        return 1
    else:
        ql.mem.write(attributeBuffer, attr)
        set_eflags_cf(ql, 0x0)
        return KERN_SUCCESS

# 0xc2
# struct rlimit {
#     rlim_t	rlim_cur;		/* current (soft) limit */       uint64
#     rlim_t	rlim_max;		/* maximum value for rlim_cur */ uint64
# };
def ql_syscall_getrlimit(ql, which, rlp, *args, **kw):
    ql.log.debug("getrlimit(which:0x%x, rlp:0x%x)" % (which, rlp))
    _RLIMIT_POSIX_FLAG = 0x1000
    RLIM_NLIMITS = 9
    which = which & _RLIMIT_POSIX_FLAG
    if which >= RLIM_NLIMITS:
        return EINVAL
    else:
        ql.mem.write(rlp, b'\x00\x13\x00\x00\x00\x00\x00\x00')  # rlim_cur
        ql.mem.write(rlp, b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x7F')  # rlim_max
        pass
    pass

# 0xc5
# this is ugly patch, we might need to get value from elf parse,
# is32bit or is64bit value not by arch
def ql_syscall_mmap2_macos(ql, mmap2_addr, mmap2_length, mmap2_prot, mmap2_flags, mmap2_fd, mmap2_pgoffset):
    MAP_ANONYMOUS=32

    if (ql.archtype== QL_ARCH.ARM64) or (ql.archtype== QL_ARCH.X8664):
        mmap2_fd = ql.unpack64(ql.pack64(mmap2_fd))

    elif (ql.archtype== QL_ARCH.MIPS):
        mmap2_fd = ql.unpack32s(ql.mem.read(mmap2_fd, 4))
        mmap2_pgoffset = ql.unpack32(ql.mem.read(mmap2_pgoffset, 4)) * 4096
        MAP_ANONYMOUS=2048
    else:
        mmap2_fd = ql.unpack32s(ql.pack32(mmap2_fd))
        mmap2_pgoffset = mmap2_pgoffset * 4096

    mmap_base = mmap2_addr
    need_mmap = True

    if mmap2_addr != 0 and mmap2_addr < ql.loader.mmap_address:
        need_mmap = False
    if mmap2_addr == 0:
        mmap_base = ql.loader.mmap_address
        ql.loader.mmap_address = mmap_base + ((mmap2_length + 0x1000 - 1) // 0x1000) * 0x1000

    ql.log.debug("mmap - mmap2(0x%x, %d, 0x%x, 0x%x, %d, %d)" % (mmap2_addr, mmap2_length, mmap2_prot, mmap2_flags, mmap2_fd, mmap2_pgoffset))
    ql.log.debug("mmap - return addr : " + hex(mmap_base))
    ql.log.debug("mmap - addr range  : " + hex(mmap_base) + ' - ' + hex(mmap_base + ((mmap2_length + 0x1000 - 1) // 0x1000) * 0x1000))

    if need_mmap:
        ql.log.debug("mmap - mapping needed")
        try:
            ql.mem.map(mmap_base, ((mmap2_length + 0x1000 - 1) // 0x1000) * 0x1000)
        except:
            pass

    ql.mem.write(mmap_base, b'\x00' * (((mmap2_length + 0x1000 - 1) // 0x1000) * 0x1000))
    
    if ((mmap2_flags & MAP_ANONYMOUS) == 0) and mmap2_fd < 256 and ql.os.fd[mmap2_fd] != 0:
        ql.os.fd[mmap2_fd].lseek(mmap2_pgoffset)
        data = ql.os.fd[mmap2_fd].read(mmap2_length)

        ql.log.debug("mem wirte : " + hex(len(data)))
        ql.log.debug("mem mmap  : " + str(ql.os.fd[mmap2_fd].name))
        ql.mem.write(mmap_base, data)
        
        mem_info = ql.os.fd[mmap2_fd].name

    ql.log.debug("mmap2(0x%x, %d, 0x%x, 0x%x, %d, %d) = 0x%x" % (mmap2_addr, mmap2_length, mmap2_prot, mmap2_flags, mmap2_fd, mmap2_pgoffset, mmap_base))
    
    regreturn = mmap_base
    ql.log.debug("mmap_base is 0x%x" % regreturn)

    return regreturn

# 0xca
def ql_syscall_sysctl(ql, name, namelen, old, oldlenp, new_arg, newlen):
    ql.log.debug("sysctl(name: 0x%x, namelen: 0x%x, old: 0x%x, oldlenp: 0x%x, new: 0x%x, newlen: 0x%x)" % (
        name, namelen, old, oldlenp, new_arg, newlen
    ))
    return KERN_SUCCESS

# 0x112
def ql_syscall_sysctlbyname(ql, name, namelen, old, oldlenp, new_arg, newlen):
    ql.log.debug("sysctlbyname(name: 0x%x, namelen: 0x%x, old: 0x%x, oldlenp: 0x%x, new: 0x%x, newlen: 0x%x)" % (
        name, namelen, old, oldlenp, new_arg, newlen
    ))
    return KERN_SUCCESS

# 0x126
# check shared region if avalible , return not ready every time
def ql_syscall_shared_region_check_np(ql, p, uap, retvalp, *args, **kw):
    ql.log.debug("shared_region_check_np(p: 0x%x, uap: 0x%x, retvalp: 0x%x) = 0x%x" % (p, uap, retvalp, EINVAL))
    return EINVAL

# 0x150
def ql_syscall_proc_info(ql, callnum, pid, flavor, arg, buff, buffer_size):
    retval = struct.unpack("<Q", ql.mem.read(ql.reg.rsp, 8))[0]
    ql.log.debug("proc_info(callnum: 0x%x, pid: %d, flavor:0x%x, arg: 0x%x, buffer: 0x%x, buffersize: 0x%x, retval: 0x%x)" % (
        callnum, pid, flavor, arg, buff, buffer_size, retval
    ))
    if callnum == PROC_INFO_CALL_PIDINFO:
        if flavor == PROC_PIDREGIONPATHINFO:
            info = ProcRegionWithPathInfo(ql)
            info.set_path(b"/usr/lib/dyld")
            info.write_info(buff)
        pass
    pass

# 0x152
def ql_syscall_stat64_macos(ql, stat64_pathname, stat64_buf_ptr, *args, **kw):
    stat64_file = (ql.mem.string(stat64_pathname))

    real_path = ql.os.macho_fs.vm_to_real_path(stat64_file)
    ql.log.debug("real_path: %s" % (real_path))
    if os.path.exists(real_path) == False:
        regreturn = -1
    else:
        stat64_info = os.stat(real_path)
        stat64_buf = ql.pack32(stat64_info.st_dev)              # st_dev            32byte
        stat64_buf += ql.pack32(stat64_info.st_mode)            # st_mode           16(32)byte
        stat64_buf += ql.pack32(stat64_info.st_nlink)           # st_nlink          16(32)byte
        stat64_buf += ql.pack64(stat64_info.st_ino)             # st_ino            64 byte
        stat64_buf += ql.pack32(0x0)                            # st_uid            32 byte
        stat64_buf += ql.pack32(0x0)                            # st_gid            32 byte
        stat64_buf += ql.pack32(0x0)                            # st_rdev           32 byte
        stat64_buf += ql.pack64(int(stat64_info.st_atime))      # st_atime          64 byte
        stat64_buf += ql.pack64(0x0)                            # st_atimensec      64 byte
        stat64_buf += ql.pack64(int(stat64_info.st_mtime))      # st_mtime          64 byte
        stat64_buf += ql.pack64(0x0)                            # st_mtimensec      64 byte
        stat64_buf += ql.pack64(int(stat64_info.st_ctime))      # st_ctime          64 byte
        stat64_buf += ql.pack64(0x0)                            # st_ctimensec      64 byte
        if ql.platform == QL_OS.MACOS:
            stat64_buf += ql.pack64(int(stat64_info.st_birthtime))  # st_birthtime      64 byte
        else:
            stat64_buf += ql.pack64(int(stat64_info.st_ctime))  # st_birthtime      64 byte
        stat64_buf += ql.pack64(0x0)                            # st_birthtimensec  64 byte
        stat64_buf += ql.pack64(stat64_info.st_size)            # st_size           64 byte
        stat64_buf += ql.pack64(stat64_info.st_blocks)          # st_blocks         64 byte
        stat64_buf += ql.pack32(stat64_info.st_blksize)         # st_blksize        32 byte
        if ql.platform == QL_OS.MACOS:
            stat64_buf += ql.pack32(stat64_info.st_flags)       # st_flags          32 byte
        else:    
            stat64_buf += ql.pack32(0x0)          
        if ql.platform == QL_OS.MACOS:
            stat64_buf += ql.pack32(stat64_info.st_gen)         # st_gen            32 byte
        else:    
            stat64_buf += ql.pack32(0x0)
        stat64_buf += ql.pack32(0x0)                            # st_lspare         32 byte
        stat64_buf += ql.pack64(0x0)                            # st_qspare         64 byte

        ql.mem.write(stat64_buf_ptr, stat64_buf)
        regreturn = 0
    ql.log.debug("stat64(%s, 0x%x) = %d" % (stat64_file, stat64_buf_ptr, regreturn))
    if regreturn == 0:
        set_eflags_cf(ql, 0x0)
        ql.log.debug("stat64 write completed")
    else:
        ql.log.debug("stat64 read/write fail")
    return regreturn

# 0x153
def ql_syscall_fstat64_macos(ql, fstat64_fd, fstat64_add, *args, **kw):
    fstat64_buf = b''
    if fstat64_fd < 256 and ql.os.fd[fstat64_fd] != 0:
        user_fileno = fstat64_fd
        fstat64_info = ql.os.fd[user_fileno].fstat()
        
        if ql.archtype== QL_ARCH.ARM64:
            fstat64_buf = ql.pack64(fstat64_info.st_dev)
            fstat64_buf += ql.pack64(fstat64_info.st_ino)
            fstat64_buf += ql.pack32(fstat64_info.st_mode)
            fstat64_buf += ql.pack32(fstat64_info.st_nlink)
            fstat64_buf += ql.pack32(1000)
            fstat64_buf += ql.pack32(1000)
            fstat64_buf += ql.pack64(fstat64_info.st_rdev)
            fstat64_buf += ql.pack64(0)
            fstat64_buf += ql.pack64(fstat64_info.st_size)
            fstat64_buf += ql.pack32(fstat64_info.st_blksize)
            fstat64_buf += ql.pack32(0)
            fstat64_buf += ql.pack64(fstat64_info.st_blocks)
            fstat64_buf += ql.pack64(int(fstat64_info.st_atime))
            fstat64_buf += ql.pack64(0)
            fstat64_buf += ql.pack64(int(fstat64_info.st_mtime))
            fstat64_buf += ql.pack64(0)
            fstat64_buf += ql.pack64(int(fstat64_info.st_ctime))
            fstat64_buf += ql.pack64(0)
        elif ql.archtype == QL_ARCH.X8664:
            # struct user64_stat64 
            fstat64_buf += ql.pack32(fstat64_info.st_dev)                   # dev_t	 	st_dev
            fstat64_buf += ql.pack32(fstat64_info.st_mode)                  # mode_t	 	st_mode
            fstat64_buf += ql.pack32(fstat64_info.st_nlink)                 # nlink_t		st_nlink
            #fstat64_buf += ql.pack32(fstat64_info.st_ino)                  # ino_t	  	st_ino
            fstat64_buf += ql.pack32(0x0)                                   # TODO: test ino_t	  	st_ino with 0x0
            fstat64_buf += ql.pack32(fstat64_info.st_uid)                   # uid_t		st_uid
            fstat64_buf += ql.pack32(fstat64_info.st_gid)                   # gid_t		st_gid
            fstat64_buf += ql.pack32(0x8800)                                # dev_t		st_rdev
            fstat64_buf += ql.pack32(int(fstat64_info.st_atime))            # user64_time_t	st_atime
            fstat64_buf += ql.pack32(0x0)                                   # user64_long_t	st_atimensec
            fstat64_buf += ql.pack32(int(fstat64_info.st_mtime))            # user64_time_t	st_mtime
            fstat64_buf += ql.pack32(0x0)                                   # user64_long_t	st_mtimensec
            fstat64_buf += ql.pack32(int(fstat64_info.st_ctime))            # user64_time_t	st_ctime
            fstat64_buf += ql.pack32(0x0)                                   # user64_long_t	st_ctimensec
            fstat64_buf += ql.pack32(0x0)                                   # user64_time_t	st_birthtime
            fstat64_buf += ql.pack32(0x0)                                   # user64_long_t	st_birthtimesec
            fstat64_buf += ql.pack32(fstat64_info.st_size)                  # off_t		st_size
            fstat64_buf += ql.pack32(0x0)                                   # blkcnt_t	st_blocks
            fstat64_buf += ql.pack32(0x0)                                   # blksize_t	st_blksize
            fstat64_buf += ql.pack32(0x0)                                   # __uint32_t	st_flags
            fstat64_buf += ql.pack32(0x0)                                   # __uint32_t	st_gen
            fstat64_buf += ql.pack32(0x0)                                   # __int32_t	st_lspare
            fstat64_buf += ql.pack32(0x0)                                   # __int64_t	st_qspare[2]
        else:
            raise QlErrorArch("Arch not support in syscall fstat64")

        ql.mem.write(fstat64_add, fstat64_buf)
        regreturn = 0
    else:
        regreturn = -1

    ql.log.debug("fstat64(%d, 0x%x) = %d" % (fstat64_fd, fstat64_add, regreturn))
    if regreturn == 0:
        ql.log.debug("fstat64 write completed")
    else:
        ql.log.debug("fstat64 read/write fail")
    return regreturn

# 0x16e
def ql_syscall_bsdthread_register(ql, threadstart, wqthread, flags, stack_addr_hint, targetconc_ptr, dispatchqueue_offset):
    set_eflags_cf(ql, 0x0)
    return 0x00000000400000df

# 0x174
def ql_syscall_thread_selfid(ql, *args, **kw):
    thread_id = ql.os.macho_thread.id
    return thread_id


# 0x18d
def ql_syscall_write_nocancel(ql, write_fd, write_buf, write_count, *args, **kw):
    regreturn = 0
    buf = None

    try:
        buf = ql.mem.read(write_buf, write_count)
        if buf:
            ql.log.debug("write() CONTENT:")
            ql.log.debug("%s" % buf)

        if hasattr(ql.os.fd[write_fd], "write"):
            ql.os.fd[write_fd].write(buf)
        else:
            ql.log.debug("write(%d,%x,%i) failed due to write_fd" % (write_fd, write_buf, write_count, regreturn))
        regreturn = write_count

    except:
        regreturn = -1

        if ql.output in (QL_OUTPUT.DEBUG, QL_OUTPUT.DUMP):
            raise
    #if buf:
    #    ql.log.info(buf.decode(errors='ignore'))
    return 0


# 0x18e
def ql_syscall_open_nocancel(ql, filename, flags, mode, *args, **kw):
    path = ql.mem.string(filename)
    relative_path = ql.os.transform_to_relative_path(path)

    flags = flags & 0xffffffff
    mode = mode & 0xffffffff

    for i in range(256):
        if ql.os.fd[i] == 0:
            idx = i
            break

    if idx == -1:
        regreturn = -1
    else:
        try:
            if ql.archtype== QL_ARCH.ARM:
                mode = 0

            ql.os.fd[idx] = ql.os.fs_mapper.open_ql_file(path, flags, mode)
            regreturn = idx
        except QlSyscallError:
            regreturn = -1

    if regreturn >= 0 and regreturn != 2:
        ql.log.debug("File Found: %s" % relative_path)
    else:
        ql.log.debug("File Not Found %s" % relative_path)
    return regreturn

# 0x1b6
def ql_syscall_shared_region_map_and_slide_np(ql, fd, count, mappings_addr, slide, slide_start, slide_size):
    ql.log.debug("shared_region_map_and_slide_np(fd: %d, count: 0x%x, mappings: 0x%x, slide: 0x%x, slide_start: 0x%x, slide_size: 0x%x)" % (
                fd, count ,mappings_addr, slide, slide_start, slide_size
            ))
    mapping_list = []
    for i in range(count):
        mapping = SharedFileMappingNp(ql)
        mapping.read_mapping(mappings_addr)
        ql.os.fd[fd].lseek(mapping.sfm_file_offset)
        content = ql.os.fd[fd].read(mapping.sfm_size)
        ql.mem.write(mapping.sfm_address, content)
        mappings_addr += mapping.size
        mapping_list.append(mapping)
    return slide_size

# 0x1e3
def ql_syscall_csrctl(ql, op, useraddr, usersize, *args, **kw):
    ql.log.debug("csrctl(op: 0x%x, useraddr :0x%x, usersize: 0x%x)" % (op, useraddr, usersize))
    return 1

# 0x1f4
def ql_syscall_getentropy(ql, buffer, size, *args, **kw):
    ql.log.debug("getentropy(buffer: 0x%x, size: 0x%x)" % (buffer, size))
    return KERN_SUCCESS

# 0x208
def ql_syscall_terminate_with_payload(ql, pid, reason_namespace, reason_code, payload, payload_size, reason_string):
    ql.log.debug("terminate_with_payload(pid: %d, reason_namespace: 0x%x, reason_code: 0x%x, payload: 0x%x \
            payload_size: 0x%x, reason_string: 0x%x)" % (pid, reason_namespace, reason_code,
            payload, payload_size, reason_string))
    ql.emu_stop()
    raise QlErrorSyscallError("Exit with Error")
    return KERN_SUCCESS

# 0x209
def ql_syscall_abort_with_payload(ql, reason_namespace, reason_code, payload, payload_size, reason_string, reason_flags):
    ql.log.debug("abort_with_payload(reason_namespace: 0x%x, reason_code: 0x%x, payload: 0x%x, payload_size: 0x%x, reason_string: 0x%x,\
            reason_flags: 0x%x)" % (reason_namespace, reason_code, payload, payload_size, reason_string, reason_flags))
    return KERN_SUCCESS



################
# mdep syscall #
################

# 0x3d
# thread_set_tsd_base
def ql_syscall_thread_fast_set_cthread_self64(ql, u_info_addr, *args, **kw):
    ql.log.debug("[mdep] thread fast set cthread self64(tsd_base:0x%x)" % (u_info_addr))
    ql.reg.msr(GSMSR, u_info_addr)
    return KERN_SUCCESS



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import struct
import time
from qiling.os.windows.const import *
from qiling.os.const import *
from qiling.os.windows.fncc import *
from qiling.os.windows.utils import *
from qiling.os.windows.thread import *
from qiling.os.windows.handle import *
from qiling.exception import *


dllname = 'kernel32_dll'

# BOOL GetStringTypeW(
#   DWORD                         dwInfoType,
#   _In_NLS_string_(cchSrc)LPCWCH lpSrcStr,
#   int                           cchSrc,
#   LPWORD                        lpCharType
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetStringTypeW(ql, address, params):
    # TODO implement
    ret = 1
    return ret


#  BOOL GetStringTypeExA
#  (
#   LCID   locale,
#   DWORD  type,
#   LPCSTR src,
#   INT    count,
#   LPWORD chartype
#  )
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetStringTypeExA(ql, address, params):
    # TODO implement
    ret = 1
    return ret


# int WideCharToMultiByte(
#   UINT                               CodePage,
#   DWORD                              dwFlags,
#   _In_NLS_string_(cchWideChar)LPCWCH lpWideCharStr,
#   int                                cchWideChar,
#   LPSTR                              lpMultiByteStr,
#   int                                cbMultiByte,
#   LPCCH                              lpDefaultChar,
#   LPBOOL                             lpUsedDefaultChar
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'LPCWCH': 'WSTRING'})
def hook_WideCharToMultiByte(ql, address, params):
    ret = 0

    cbMultiByte = params["cbMultiByte"]
    s_lpWideCharStr = params["lpWideCharStr"]
    lpMultiByteStr = params["lpMultiByteStr"]
    s = (s_lpWideCharStr + "\x00").encode("utf-16le")
    if cbMultiByte != 0 and lpMultiByteStr != 0:
        ql.mem.write(lpMultiByteStr, s)
    ret = len(s)

    return ret


# int MultiByteToWideChar(
#  UINT                              CodePage,
#  DWORD                             dwFlags,
#  _In_NLS_string_(cbMultiByte)LPCCH lpMultiByteStr,
#  int                               cbMultiByte,
#  LPWSTR                            lpWideCharStr,
#  int                               cchWideChar
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'DWORD': 'UINT', 'LPCCH': 'WSTRING'})
def hook_MultiByteToWideChar(ql, address, params):
    wide_str = (params['lpMultiByteStr']+"\x00").encode('utf-16le')
    if params['cchWideChar'] != 0:
        ql.mem.write(params['lpWideCharStr'], wide_str)
    return len(wide_str)



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *

def ql_syscall__sysctl(ql, sysctl_name, sysctl_namelen, sysctl_bytes_oldlenp, sysctl_size_oldlenp, sysctl_bytes_newlen, sysctl_size_newlen):
    # sysctl (name=0x7fffffffe3d8, namelen=2, oldp=0x7fffffffe3d4, oldlenp=0x7fffffffe3e0, newp=0x0, newlen=<optimized out>)
    regreturn = 0
    return regreturn



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from struct import unpack

from .const import *
from .utils import *

# TODO: We need support more LC command to load more kinds of binary

class LoadCommand:

    def __init__(self, data):
        self.data = data
        self.FR = FileReader(data)
        self.cmd_id = unpack("<L", self.FR.read(4))[0]
        self.cmd_size = unpack("<L", self.FR.read(4))[0]
        pass

    def get_complete(self):
        cmd_map = {
            LC_SEGMENT_64           :   LoadSegment64,
            LC_SEGMENT              :   LoadSegment32,
            LC_SYMTAB               :   LoadSymtab,
            LC_DYSYMTAB             :   LoadDysymtab,
            LC_ID_DYLINKER          :   LoadIdDylinker,
            LC_UUID                 :   LoadUuid,
            LC_VERSION_MIN_MACOSX   :   LoadVersionMinMacosx,
            LC_VERSION_MIN_IPHONEOS :   LoadVersionMinIphoneos,
            LC_SOURCE_VERSION       :   LoadSourceVersion,
            LC_UNIXTHREAD           :   LoadUnixThread,
            LC_SEGMENT_SPLIT_INFO   :   LoadSegmentSplitInfo,
            LC_FUNCTION_STARTS      :   LoadFunctionStarts,
            LC_DATA_IN_CODE         :   LoadDataInCode,
            LC_CODE_SIGNATURE       :   LoadCodeSignature,
            LC_DYLD_INFO_ONLY       :   LoadDyldInfoOnly,
            LC_LOAD_DYLINKER        :   LoadDylinker,
            LC_MAIN                 :   LoadMain,
            LC_LOAD_DYLIB           :   LoadDyLib,
            LC_ENCRYPTION_INFO_64   :   LoadEncryptionInfo64,
            LC_DYLD_EXPORTS_TRIE    :   LoadDyldExportTrie,
            LC_DYLD_CHAINED_FIXUPS  :   LoadDyldChainedFixups,
            LC_BUILD_VERSION        :   LoadBuildVersion
        }

        exec_func = cmd_map.get(self.cmd_id)
        if exec_func:
            return exec_func(self.data)
    

class LoadSegment(LoadCommand):
    
    def __init__(self, data):
        super().__init__(data)
        self.segment_name = self.FR.read(0x10).decode("utf-8")

    def get_complete(self):
        pass


class LoadSegment32(LoadSegment):
    
    def __init__(self, data):
        super().__init__(data)
        self.vm_address             = unpack("<L", self.FR.read(4))[0]
        self.vm_size                = unpack("<L", self.FR.read(4))[0]
        self.file_offset            = unpack("<L", self.FR.read(4))[0]
        self.file_size              = unpack("<L", self.FR.read(4))[0]
        self.maximum_vm_protection  = unpack("<L", self.FR.read(4))[0]
        self.initial_vm_protection  = unpack("<L", self.FR.read(4))[0]
        self.number_of_sections     = unpack("<L", self.FR.read(4))[0]
        self.flags                  = unpack("<L", self.FR.read(4))[0]
        self.sections               = []
        if self.number_of_sections:
            for i in range(self.number_of_sections):
                self.sections.append(LoadSection32(self.FR.read(0x44)))

    def get_complete(self):
        pass


class LoadSegment64(LoadSegment):
    # FIXME: segmengs should not be fixed size of 0x1000, should be calculated
    # header mark as 0x1000 but it seems we need count it to matched IDApro
    def __init__(self, data):
        super().__init__(data)
        self.vm_address             = unpack("<Q", self.FR.read(8))[0]
        self.vm_size                = unpack("<Q", self.FR.read(8))[0]
        self.file_offset            = unpack("<Q", self.FR.read(8))[0]
        self.file_size              = unpack("<Q", self.FR.read(8))[0]
        self.maximum_vm_protection  = unpack("<L", self.FR.read(4))[0]
        self.initial_vm_protection  = unpack("<L", self.FR.read(4))[0]
        self.number_of_sections     = unpack("<L", self.FR.read(4))[0]
        self.flags                  = unpack("<L", self.FR.read(4))[0]
        self.sections               = []
        if self.number_of_sections:
            for i in range(self.number_of_sections):
                self.sections.append(LoadSection64(self.FR.read(0x50)))
        
    # def __str__(self):
    #     return (" SEG64:Seg Name %s, vmaddr 0x%X, vm size 0x%X, file offset 0x%X, file size 0x%X, max vp 0x%X, init vp 0x%X, section num %d, flags 0x%X" % (
    #        self.segment_name, self.vm_address, self.vm_size, self.file_offset, self.file_size, self.maximum_vm_protection,
    #        self.initial_vm_protection, self.number_of_sections, self.flags 
    #     ))

    def get_complete(self):
        pass


class LoadSymtab(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.symbol_table_offset    = unpack("<L", self.FR.read(4))[0]
        self.number_of_symbols      = unpack("<L", self.FR.read(4))[0]
        self.string_table_offset    = unpack("<L", self.FR.read(4))[0]
        self.string_table_size      = unpack("<L", self.FR.read(4))[0]


    # def __str__(self):
    #     return (" Symtab: sym table offset 0x%X, sym num 0x%X, str table offset 0x%X, str table size 0x%X" % (
    #         self.symbol_table_offset, self.number_of_symbols, self.string_table_offset, self.string_table_size
    #     ))

    def get_complete(self):
        pass


class LoadDysymtab(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.locsymbol_index            = unpack("<L", self.FR.read(4))[0]
        self.locsymbol_number           = unpack("<L", self.FR.read(4))[0]
        self.defined_extsymbol_index    = unpack("<L", self.FR.read(4))[0]
        self.defined_extsymbol_number   = unpack("<L", self.FR.read(4))[0]
        self.undef_extsymbol_index      = unpack("<L", self.FR.read(4))[0]
        self.undef_extsymbol_number     = unpack("<L", self.FR.read(4))[0]
        self.toc_offset                 = unpack("<L", self.FR.read(4))[0]
        self.toc_entries                = unpack("<L", self.FR.read(4))[0]
        self.module_table_offset        = unpack("<L", self.FR.read(4))[0]
        self.module_table_entries       = unpack("<L", self.FR.read(4))[0]
        self.extref_table_offset        = unpack("<L", self.FR.read(4))[0]
        self.extref_table_entries       = unpack("<L", self.FR.read(4))[0]
        self.indsym_table_offset        = unpack("<L", self.FR.read(4))[0]
        self.indsym_table_entries       = unpack("<L", self.FR.read(4))[0]
        self.extreloc_table_offset      = unpack("<L", self.FR.read(4))[0]
        self.extreloc_table_entries     = unpack("<L", self.FR.read(4))[0]
        self.locreloc_table_offset      = unpack("<L", self.FR.read(4))[0]
        self.locreloc_table_entries     = unpack("<L", self.FR.read(4))[0]

    def get_complete(self):
        pass


class LoadDylinker(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.str_offset = unpack("<L", self.FR.read(4))[0]
        self.name = self.FR.readString(4)

    # def __str__(self):
    #     return (" DyLinker: Name %s" % self.name)

    def get_complete(self):
        pass


class LoadIdDylinker(LoadCommand):

    def __init__(self, data):
        super().__init__(data)

    def get_complete(self):
        pass


class LoadUuid(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.uuid = self.FR.read(16)

    def get_complete(self):
        pass


class LoadVersionMinMacosx(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.version    = unpack("<L", self.FR.read(4))[0]
        self.reserved   = unpack("<L", self.FR.read(4))[0]

    # def __str__(self):
    #     return (" VersionMinMacosx: version %X" % self.version)

    def get_complete(self):
        pass

class LoadVersionMinIphoneos(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.version    = unpack("<L", self.FR.read(4))[0]
        self.reserved   = unpack("<L", self.FR.read(4))[0]

    # def __str__(self):
    #     return (" VersionMinIphoneos: version %X" % self.version)

    def get_complete(self):
        pass


class LoadSourceVersion(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.version = unpack("<Q", self.FR.read(8))[0]

    def get_complete(self):
        pass


class LoadMain(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.entry_offset   = unpack("<Q", self.FR.read(8))[0]
        self.stack_size     = unpack("<Q", self.FR.read(8))[0]

    def get_complete(self):
        pass


class LoadDyLib(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.str_offset             = unpack("<L", self.FR.read(4))[0]
        self.time_stamp             = unpack("<L", self.FR.read(4))[0]
        self.current_version        = unpack("<L", self.FR.read(4))[0]
        self.compatibility_version  = unpack("<L", self.FR.read(4))[0]
        self.name                   = self.FR.readString(4)
    
    # def __str__(self):
    #     return (" Dylib: name %s" % self.name)

    def get_complete(self):
        pass


class LoadUnixThread(LoadCommand):

    def __init__(self, data):
        super().__init__(data)

        self.flavor = unpack("<L", self.FR.read(4))[0]
        self.count  = unpack("<L", self.FR.read(4))[0]
        
        if self.flavor == X86_THREAD_STATE32:
            self.eax        = unpack("<L", self.FR.read(4))[0]
            self.ebx        = unpack("<L", self.FR.read(4))[0]
            self.ecx        = unpack("<L", self.FR.read(4))[0]
            self.edx        = unpack("<L", self.FR.read(4))[0]
            self.ebi        = unpack("<L", self.FR.read(4))[0]
            self.esi        = unpack("<L", self.FR.read(4))[0]
            self.ebp        = unpack("<L", self.FR.read(4))[0]
            self.esp        = unpack("<L", self.FR.read(4))[0]
            self.ss         = unpack("<L", self.FR.read(4))[0]
            self.eflags     = unpack("<L", self.FR.read(4))[0]
            self.eip        = unpack("<L", self.FR.read(4))[0]
            self.cs         = unpack("<L", self.FR.read(4))[0]
            self.ds         = unpack("<L", self.FR.read(4))[0]
            self.es         = unpack("<L", self.FR.read(4))[0]
            self.fs         = unpack("<L", self.FR.read(4))[0]
            self.gs         = unpack("<L", self.FR.read(4))[0]
            self.entry      = self.eip
        
        elif self.flavor == X86_THREAD_STATE64:
            self.rax        = unpack("<Q", self.FR.read(8))[0]
            self.rbx        = unpack("<Q", self.FR.read(8))[0]
            self.rcx        = unpack("<Q", self.FR.read(8))[0]
            self.rdx        = unpack("<Q", self.FR.read(8))[0]
            self.rdi        = unpack("<Q", self.FR.read(8))[0]
            self.rsi        = unpack("<Q", self.FR.read(8))[0]
            self.rbp        = unpack("<Q", self.FR.read(8))[0]
            self.rsp        = unpack("<Q", self.FR.read(8))[0]
            self.r8         = unpack("<Q", self.FR.read(8))[0]
            self.r9         = unpack("<Q", self.FR.read(8))[0]
            self.r10        = unpack("<Q", self.FR.read(8))[0]
            self.r11        = unpack("<Q", self.FR.read(8))[0]
            self.r12        = unpack("<Q", self.FR.read(8))[0]
            self.r13        = unpack("<Q", self.FR.read(8))[0]
            self.r14        = unpack("<Q", self.FR.read(8))[0]
            self.r15        = unpack("<Q", self.FR.read(8))[0]
            self.rip        = unpack("<Q", self.FR.read(8))[0]
            self.rflags     = unpack("<Q", self.FR.read(8))[0]
            self.cs         = unpack("<Q", self.FR.read(8))[0]
            self.fs         = unpack("<Q", self.FR.read(8))[0]
            self.gs         = unpack("<Q", self.FR.read(8))[0]
            self.entry      = self.rip
        
        elif self.flavor == ARM_THREAD_STATE64:
            self.x0         = unpack("<Q", self.FR.read(8))[0]
            self.x1         = unpack("<Q", self.FR.read(8))[0]
            self.x2         = unpack("<Q", self.FR.read(8))[0]
            self.x3         = unpack("<Q", self.FR.read(8))[0]
            self.x4         = unpack("<Q", self.FR.read(8))[0]
            self.x5         = unpack("<Q", self.FR.read(8))[0]
            self.x6         = unpack("<Q", self.FR.read(8))[0]
            self.x7         = unpack("<Q", self.FR.read(8))[0]
            self.x8         = unpack("<Q", self.FR.read(8))[0]
            self.x9         = unpack("<Q", self.FR.read(8))[0]
            self.x10        = unpack("<Q", self.FR.read(8))[0]
            self.x11        = unpack("<Q", self.FR.read(8))[0]
            self.x12        = unpack("<Q", self.FR.read(8))[0]
            self.x13        = unpack("<Q", self.FR.read(8))[0]
            self.x14        = unpack("<Q", self.FR.read(8))[0]
            self.x15        = unpack("<Q", self.FR.read(8))[0]
            self.x16        = unpack("<Q", self.FR.read(8))[0]
            self.x17        = unpack("<Q", self.FR.read(8))[0]
            self.x18        = unpack("<Q", self.FR.read(8))[0]
            self.x19        = unpack("<Q", self.FR.read(8))[0]
            self.x20        = unpack("<Q", self.FR.read(8))[0]
            self.x21        = unpack("<Q", self.FR.read(8))[0]            
            self.r22        = unpack("<Q", self.FR.read(8))[0]
            self.x23        = unpack("<Q", self.FR.read(8))[0]
            self.x24        = unpack("<Q", self.FR.read(8))[0]
            self.x25        = unpack("<Q", self.FR.read(8))[0]
            self.x26        = unpack("<Q", self.FR.read(8))[0]
            self.x27        = unpack("<Q", self.FR.read(8))[0]            
            self.x28        = unpack("<Q", self.FR.read(8))[0]
            self.x29        = unpack("<Q", self.FR.read(8))[0]
            self.x30        = unpack("<Q", self.FR.read(8))[0]            
            self.sp         = unpack("<Q", self.FR.read(8))[0]            
            self.pc         = unpack("<Q", self.FR.read(8))[0]
            self.entry      = self.pc

    #def __str__(self):
    #    return (" Unixthread: entry 0x%X" %self.entry)  

    def get_complete(self):
        pass


class LoadSegmentSplitInfo(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.data_offset    = unpack("<L", self.FR.read(4))[0]
        self.data_size      = unpack("<L", self.FR.read(4))[0]

    def get_complete(self):
        pass


class LoadFunctionStarts(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.data_offset    = unpack("<L", self.FR.read(4))[0]
        self.data_size      = unpack("<L", self.FR.read(4))[0]

    def get_complete(self):
        pass


class LoadDataInCode(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.data_offset    = unpack("<L", self.FR.read(4))[0]
        self.data_size      = unpack("<L", self.FR.read(4))[0]

    # def __str__(self):
    #     return (" Data in code: offset 0x%X" % self.data_offset)

    def get_complete(self):
        pass


class LoadCodeSignature(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.data_offset    = unpack("<L", self.FR.read(4))[0]
        self.data_size      = unpack("<L", self.FR.read(4))[0]

    # def __str__(self):
    #     return (" CodeSignature: offset 0x%X" % self.data_offset)

    def get_complete(self):
        pass


class LoadDyldInfoOnly(LoadCommand):

    def __init__(self, data):
        super().__init__(data)
        self.rebase_info_offset         = unpack("<L", self.FR.read(4))[0]
        self.rebase_info_size           = unpack("<L", self.FR.read(4))[0]
        self.binding_info_offset        = unpack("<L", self.FR.read(4))[0]
        self.binding_info_size          = unpack("<L", self.FR.read(4))[0]
        self.weak_binding_info_offset   = unpack("<L", self.FR.read(4))[0]
        self.weak_binding_info_size     = unpack("<L", self.FR.read(4))[0]
        self.lazy_binding_info_offset   = unpack("<L", self.FR.read(4))[0]
        self.lazy_binding_info_size     = unpack("<L", self.FR.read(4))[0]
        self.export_info_offset         = unpack("<L", self.FR.read(4))[0]
        self.export_info_size           = unpack("<L", self.FR.read(4))[0]

#     def __str__(self):
#         return (" DyldInfoOnly: rebase offset: 0x%X, rebase size 0x%X, binding offset 0x%X, binding size 0x%X,\
#  weak offset 0x%X, weak size 0x%X, lazy offset 0x%X, lazy size 0x%X, export offset 0x%X, export size 0x%X" %(
#             self.rebase_info_offset, self.rebase_info_size, self.binding_info_offset, self.binding_info_size,
#             self.weak_binding_info_offset, self.weak_binding_info_size, self.lazy_binding_info_offset,
#             self.lazy_binding_info_size, self.export_info_offset, self.export_info_size 
#         ))

    def get_complete(self):
        pass


class LoadSection:

    def __init__(self, data):
        self.FR = FileReader(data)
        self.section_name   = self.FR.read(0x10).decode("utf-8")
        self.segment_name   = self.FR.read(0x10).decode("utf-8")
        

class LoadSection32(LoadSection):

    # size 0x44
    def __init__(self, data):
        super().__init__(data)
        self.address                = unpack("<L", self.FR.read(4))[0]
        self.size                   = unpack("<L", self.FR.read(4))[0]
        self.offset                 = unpack("<L", self.FR.read(4))[0]
        self.alignment              = unpack("<L", self.FR.read(4))[0]
        self.relocations_offset     = unpack("<L", self.FR.read(4))[0]
        self.number_of_relocations  = unpack("<L", self.FR.read(4))[0]
        self.flags                  = unpack("<L", self.FR.read(4))[0]
        self.reserved1              = unpack("<L", self.FR.read(4))[0]
        self.reserved2              = unpack("<L", self.FR.read(4))[0]
  
    # def __str__(self):
    #     return ("     Section name %s, Seg name %s, addr 0x%X, size 0x%X, offset 0x%X, align 0x%X, rel offset 0x%X, rel num %d, flags 0x%X" % (
    #         self.section_name, self.segment_name, self.address, self.size, self.offset, self.alignment, self.relocations_offset,
    #         self.number_of_relocations, self.flags
    #     ))


class LoadSection64(LoadSection):

    # size 0x50
    def __init__(self, data):
        super().__init__(data)
        self.address                = unpack("<Q", self.FR.read(8))[0]
        self.size                   = unpack("<Q", self.FR.read(8))[0]
        self.offset                 = unpack("<L", self.FR.read(4))[0]
        self.alignment              = unpack("<L", self.FR.read(4))[0]
        self.relocations_offset     = unpack("<L", self.FR.read(4))[0]
        self.number_of_relocations  = unpack("<L", self.FR.read(4))[0]
        self.flags                  = unpack("<L", self.FR.read(4))[0]
        self.reserved1              = unpack("<L", self.FR.read(4))[0]
        self.reserved2              = unpack("<L", self.FR.read(4))[0]
        self.reserved3              = unpack("<L", self.FR.read(4))[0]
        
    # def __str__(self):
    #     return ("     Sec64: Section name %s, Seg name %s, addr 0x%X, size 0x%X, offset 0x%X, align 0x%X, rel offset 0x%X, rel num %d, flags 0x%X" % (
    #         self.section_name, self.segment_name, self.address, self.size, self.offset, self.alignment, self.relocations_offset,
    #         self.number_of_relocations, self.flags
    #     ))


class LoadEncryptionInfo64(LoadCommand):

    def __init__(self, data):
        super().__init__(data)

    def get_complete(self):
        pass        


class LoadDyldExportTrie(LoadCommand):

    def __init__(self, data):
        super().__init__(data)

    def get_complete(self):
        pass       


class LoadDyldChainedFixups(LoadCommand):

    def __init__(self, data):
        super().__init__(data)

    def get_complete(self):
        pass           


class LoadBuildVersion(LoadCommand):

    def __init__(self, data):
        super().__init__(data)

    def get_complete(self):
        pass        
    



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#
# A Simple Windows Clipboard Simulation


from unicorn import *
from unicorn.x86_const import *

from qiling.const import *

class Fiber:
    def __init__(self, idx, cb=None):
        self.idx = idx
        self.data = 0
        self.cb = cb


class FiberManager:
    def __init__(self, ql):
        self.fibers = {}
        self.idx = 0
        self.ql = ql

    def alloc(self, cb=None):
        rtn = self.idx
        self.fibers[self.idx] = Fiber(self.idx, cb=cb)
        self.idx += 1
        return rtn

    def free(self, idx):
        if idx not in self.fibers:
            self.last_error = 0x57  # ERROR_INVALID_PARAMETER
            return 0
        else:
            fiber = self.fibers[idx]
            if fiber.cb:
                self.ql.log.debug("Skipping emulation of callback function 0x%X for fiber 0x%X" % (fiber.cb, fiber.idx))
                """
                ret_addr = self.ql.reg.read(UC_X86_REG_RIP + 6 ) #FIXME, use capstone to get addr of next instr?

                # Write Fls data to memory to be accessed by cb
                addr = self.ql.os.heap.alloc(self.ql.pointersize)
                data = fiber.data.to_bytes(self.ql.pointersize, byteorder='little')
                self.ql.mem.write(addr, data)

                # set up params and return address then jump to callback
                if self.ql.pointersize == 8:
                    self.ql.reg.write(UC_X86_REG_RCX, addr)
                else:
                    self.ql.stack_push(ret_addr)
                self.ql.stack_push(ret_addr)
                self.ql.log.debug("Jumping to callback @ 0x%X" % fiber.cb)
                self.ql.reg.write(UC_X86_REG_RIP, fiber.cb)
                # All of this gets overwritten by the rest of the code in fncc.py
                # Not sure how to actually make unicorn emulate the callback function due to that
                """
            else:
                del self.fibers[idx]
                return 1

    def set(self, idx, data):
        if idx not in self.fibers:
            self.last_error = 0x57  # ERROR_INVALID_PARAMETER
            return 0
        else:
            self.fibers[idx].data = data
            return 1

    def get(self, idx):
        if idx not in self.fibers:
            self.last_error = 0x57  # ERROR_INVALID_PARAMETER
            return 0
        else:
            return self.fibers[idx].data
